# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import tempfile
from abc import ABC, abstractmethod, abstractproperty
from collections import defaultdict, namedtuple
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

from nemo.utils import logging


TW_BREAK = "‡"


try:
    import kaldifst

    # check that kaldifst package is not empty
    # Note: lightning.pytorch.utilities.imports.package_available may not help here
    kaldifst.StdVectorFst()
    _KALDIFST_AVAILABLE = True
except (ImportError, ModuleNotFoundError, AttributeError):
    _KALDIFST_AVAILABLE = False


try:
    import graphviz

    _GRAPHVIZ_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    _GRAPHVIZ_AVAILABLE = False


try:
    import kaldilm

    _KALDILM_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    _KALDILM_AVAILABLE = False


KALDIFST_INSTALLATION_MESSAGE = (
    "kaldifst is not installed or is installed incorrectly.\n"
    "please run `pip install kaldifst` or `bash scripts/installers/install_riva_decoder.sh` to install."
)


GRAPHVIZ_INSTALLATION_MESSAGE = (
    "graphviz is not installed.\n" "please run `bash scripts/installers/install_graphviz.sh` to install."
)


KALDILM_INSTALLATION_MESSAGE = (
    "kaldilm is not installed.\n"
    "please run `pip install kaldilm` or `bash scripts/installers/install_riva_decoder.sh` to install."
)


def _kaldifst_maybe_raise():
    if _KALDIFST_AVAILABLE is False:
        raise ImportError(KALDIFST_INSTALLATION_MESSAGE)


def kaldifst_importer():
    """Import helper function that returns kaldifst package or raises ImportError exception."""
    _kaldifst_maybe_raise()
    return kaldifst


def _graphviz_maybe_raise():
    if _GRAPHVIZ_AVAILABLE is False:
        raise ImportError(GRAPHVIZ_INSTALLATION_MESSAGE)


def graphviz_importer():
    """Import helper function that returns graphviz package or raises ImportError exception."""
    _graphviz_maybe_raise()
    return graphviz


def _kaldilm_maybe_raise():
    if _KALDILM_AVAILABLE is False:
        raise ImportError(KALDILM_INSTALLATION_MESSAGE)


def kaldilm_importer():
    """Import helper function that returns kaldifst package or raises ImportError exception."""
    _kaldilm_maybe_raise()
    return kaldilm


@dataclass
class LexiconUnit:
    """A dataclass encapsulating the name of the language unit (e.g. wordpiece) and its mark (e.g. word begin)."""

    name: str
    mark: str = ""


class Lexicon:
    def __init__(
        self,
        wordid2tokenid: Dict[int, List[List[int]]],
        id2word: Union[Dict[int, str], Dict[int, LexiconUnit]],
        id2token: Union[Dict[int, str], Dict[int, LexiconUnit]],
        disambig_pattern: str = re.compile(r"^#\d+$"),
    ):
        """
        Lexicon class which contains word-to-token-sequence, word-to-id, and token-to-id mappings.

        Args:
          wordid2tokenid:
            Lexicon.
            Mapping from word_id to token1_id token2_id ... tokenN_id.

          id2word:
            Word index.
            Mapping from word_id to word_str.

          id2token:
            Token index.
            Mapping from token_id to token_str.

          disambig_pattern:
            Pattern for disambiguation symbols.
        """
        is_id2token_str = not isinstance(list(id2token.values())[0], LexiconUnit)
        self.id2token = {k: LexiconUnit(v) for k, v in id2token.items()} if is_id2token_str else id2token
        self.token2id = {v.name: k for k, v in self.id2token.items()}
        is_id2word_str = not isinstance(list(id2word.values())[0], LexiconUnit)
        self.id2word = {k: LexiconUnit(v) for k, v in id2word.items()} if is_id2word_str else id2word
        self.word2id = {v.name: k for k, v in self.id2word.items()}
        self.wordid2tokenid = wordid2tokenid
        word2tokens = defaultdict(list)
        for k, v in self.wordid2tokenid.items():
            word2tokens[self.id2word[k].name] += [[self.id2token[i].name for i in vp] for vp in v]
        self.word2tokens = word2tokens
        self.disambig_pattern = disambig_pattern

        max_disambig_id = -1
        num_disambigs = 0
        self.has_epsilon = False
        self._default_disambig_mark = "disambig"
        self._default_epsilon_mark = "epsilon"
        self._default_epsilon_name = "<eps>"
        for i, s in self.id2token.items():
            if self.disambig_pattern.match(s.name):
                if is_id2token_str or not s.mark.startswith(self._default_disambig_mark):
                    s.mark = self._default_disambig_mark
                if i > max_disambig_id:
                    max_disambig_id = i
                    num_disambigs += 1
            if s.name == self._default_epsilon_name or s.mark == self._default_epsilon_mark:
                assert i == 0
                self.has_epsilon = True
        self.max_disambig_id = max_disambig_id
        self.num_disambigs = num_disambigs

        if is_id2word_str:
            for i, s in self.id2word.items():
                if self.disambig_pattern.match(s.name):
                    s.mark = self._default_disambig_mark
                elif s.name == self._default_epsilon_name:
                    s.mark == self._default_epsilon_mark

    def __iter__(self) -> Tuple[str, List[str]]:
        for wordid, tokenid_list in self.wordid2tokenid.items():
            for tokenids in tokenid_list:
                yield wordid, tokenids

    def __str__(self):
        return str(self.word2tokens)

    @property
    def token_ids(self) -> List[int]:
        """Return a list of token IDs excluding those from
        disambiguation symbols.
        """
        ans = []
        for i, s in self.id2token.items():
            if not s.mark.startswith(self._default_epsilon_mark) and (not self.has_epsilon or i != 0):
                ans.append(i)
        ans.sort()
        return ans


def arpa2fst(lm_path: str, attach_symbol_table: bool = True) -> 'kaldifst.StdVectorFst':
    """
    Compiles an ARPA LM file into a grammar WFST (G.fst).

    Args:
      lm_path:
        Path to the ARPA LM file.

      attach_symbol_table:
        Whether to attach the words for indices of the returned WFST.

    Returns:
      Kaldi-type grammar WFST.
    """
    _kaldifst_maybe_raise()
    _kaldilm_maybe_raise()

    with tempfile.TemporaryDirectory() as tempdirname:
        output_fst = os.path.join(tempdirname, "output.fst")
        words_txt = os.path.join(tempdirname, "words.txt")
        #         with suppress_stdout_stderr():
        kaldilm.arpa2fst(
            input_arpa=lm_path,
            output_fst=output_fst,
            disambig_symbol="#0",
            write_symbol_table=words_txt,
        )

        G = kaldifst.StdVectorFst.read(output_fst)

        if attach_symbol_table:
            osym = kaldifst.SymbolTable()
            with open(words_txt, encoding="utf-8") as f:
                for line in f:
                    w, i = line.strip().split()
                    osym.add_symbol(symbol=w, key=int(i))
            G.output_symbols = osym

    kaldifst.arcsort(G, sort_type="ilabel")
    return G


def add_tokenwords_(
    g_fst: 'kaldifst.StdVectorFst',
    tokens: List[str],
    word_weight: float = 2.0,
    token_unigram_weight: float = 4.0,
    token_oov: str = "<unk>",
) -> int:
    """
    Adds special words representing individual tokens (tokenwords).
    In-place operation.

    Args:
      g_fst:
        Kaldi-type grammar WFST.
        Will be augmented with the tokenwords.

      tokens:
        Token vocabulary.

      word_weight:
        The weight of an Out Of Vocabulary (OOV) word emission.

      token_unigram_weight:
        The weight of a tokenword emission.

      token_oov:
        OOV token.

    Returns:
        The id of the tokenword disambiguation token.
    """
    _kaldifst_maybe_raise()

    unigram_state = 0
    # check if 0 is the unigram state (has no outgoing epsilon arcs)
    assert kaldifst.ArcIterator(g_fst, unigram_state).value.ilabel not in (0, g_fst.output_symbols.find("#0"))

    # we put tokenword self-loops in a separate state wrapped with a tokenword_disambig token
    tokenword_disambig_id = g_fst.output_symbols.available_key()
    tokenword_disambig = "#1"
    g_fst.output_symbols.add_symbol(tokenword_disambig, tokenword_disambig_id)
    tokenword_state = g_fst.add_state()
    # we keep olabel !=0 to mark tokenword segments in the recognition results
    g_fst.add_arc(
        state=unigram_state,
        arc=kaldifst.StdArc(
            ilabel=tokenword_disambig_id,
            olabel=tokenword_disambig_id,
            weight=word_weight,
            nextstate=tokenword_state,
        ),
    )
    g_fst.add_arc(
        state=tokenword_state,
        arc=kaldifst.StdArc(
            ilabel=tokenword_disambig_id,
            olabel=tokenword_disambig_id,
            weight=0.0,
            nextstate=unigram_state,
        ),
    )
    label = tokenword_disambig_id + 1
    for t in tokens:
        if t != token_oov:
            g_fst.add_arc(
                state=tokenword_state,
                arc=kaldifst.StdArc(
                    ilabel=label,
                    olabel=label,
                    weight=token_unigram_weight,
                    nextstate=tokenword_state,
                ),
            )
            g_fst.output_symbols.add_symbol(f"{t}{TW_BREAK}{tokenword_disambig}", label)
            label += 1

    return tokenword_disambig_id


def generate_lexicon_sentencepiece(
    tokenizer: 'TokenizerSpec',
    id2word: Dict[int, str],
    oov: str = "<unk>",
    add_epsilon: bool = False,
    first_tokenword_id: int = -1,
    disambig_pattern: str = re.compile(r"^#\d+$"),
) -> Lexicon:
    """
    Generate a Lexicon using a SentencePiece tokenizer.

    Args:
      tokenizer:
        NeMo SentencePiece tokenizer.

      id2word:
        Word index.
        Mapping from word_id to word_str.

      oov:
        Out Of Vocabulary word in lexicon.

    Returns:
      Lexicon object.
    """
    word2id = {v: k for k, v in id2word.items()}
    backoff_disambig = "#0"
    tokenword_disambig = "#1"
    word_begin_mark = "▁"

    tokenword_mode = first_tokenword_id != -1
    if tokenword_mode:
        words, tokenwords = [], []
        for k, v in id2word.items():
            if disambig_pattern.match(v):
                continue
            words.append(v) if k < first_tokenword_id else tokenwords.append(v)
    else:
        words, tokenwords = [v for v in id2word.values() if not disambig_pattern.match(v)], []

    # Use encode to avoid OOV tokens
    words_piece_ids = tokenizer.encode(words, out_type=int)

    # tokenizer.get_vocab() gives indices starting with 1
    maybe_add_one = int(add_epsilon)
    maybe_subtract_one = int(not add_epsilon)
    vocab = tokenizer.get_vocab()
    id2token = {
        v - maybe_subtract_one: LexiconUnit(k, "begin" if k.startswith(word_begin_mark) else "")
        for k, v in vocab.items()
    }

    # Introduce unk, blank, and the first disambig ids
    unk_id = tokenizer.piece_to_id(oov) + maybe_add_one
    id2token[unk_id] = LexiconUnit(oov, "unk")
    # We assume blank to have the last output id of the neural network output
    max_token_id = max(id2token.keys())
    id2token[max_token_id + 1] = LexiconUnit("<blk>", "blank")
    id2token[max_token_id + 2] = LexiconUnit(backoff_disambig, "disambig_backoff")
    if tokenword_mode:
        id2token[max_token_id + 3] = LexiconUnit(tokenword_disambig, "disambig_tokenword")
    if add_epsilon:
        # insert first
        id2token[0] = LexiconUnit("<eps>", "epsilon")
        id2token = {k: v for k, v in sorted(id2token.items(), key=lambda item: item[0])}

    if tokenword_mode:
        words += tokenwords
        words_piece_ids += [[vocab[tw.rstrip(f"{TW_BREAK}{tokenword_disambig}")] - maybe_add_one] for tw in tokenwords]

    wordid2tokenid = defaultdict(list)

    for word, piece_ids in zip(words, words_piece_ids):
        if word.startswith("<") and word != "<eps>":  # not a real word, probably some tag
            continue
        elif word == "<eps>":  # we do not need to tokelize <eps>
            continue
        else:
            wordid2tokenid[word2id[word]].append([p + maybe_add_one for p in piece_ids])

    lexicon = Lexicon(wordid2tokenid, id2word, id2token)
    # state disambig purpose explicitly for further use
    lexicon.id2word[lexicon.word2id[backoff_disambig]].mark = "disambig_backoff"
    if tokenword_mode:
        lexicon.id2word[lexicon.word2id[tokenword_disambig]].mark = "disambig_tokenword"
        for tw in tokenwords:
            lexicon.id2word[lexicon.word2id[tw]].mark = "tokenword"
    return lexicon


def add_disambig_symbols(lexicon: Lexicon) -> Lexicon:
    """
    Adds pseudo-token disambiguation symbols #1, #2 and so on
    at the ends of tokens to ensure that all pronunciations are different,
    and that none is a prefix of another.

    See also add_lex_disambig.pl from kaldi.

    Args:
      lexicon:
        Lexicon object.

    Returns:
      Return Lexicon augmented with subseqence disambiguation symbols.
    """

    tokenword_mode = "#1" in lexicon.word2id
    if tokenword_mode:
        first_tokenword_id = lexicon.word2id["#1"] + 1
        last_used_disambig_id = lexicon.token2id["#1"]
    else:
        last_used_disambig_id = lexicon.token2id["#0"]

    # (1) Work out the count of each token-sequence in the lexicon.
    count = defaultdict(int)
    for _, token_ids in lexicon:
        count[tuple(token_ids)] += 1

    # (2) For each left sub-sequence of each token-sequence, note down
    # that it exists (for identifying prefixes of longer strings).
    issubseq = defaultdict(int)
    for word_id, token_ids in lexicon:
        if tokenword_mode and word_id >= first_tokenword_id:
            continue
        token_ids = token_ids.copy()
        token_ids.pop()
        while token_ids:
            issubseq[tuple(token_ids)] = 1
            token_ids.pop()

    # (3) For each entry in the lexicon:
    # if the token sequence is unique and is not a
    # prefix of another word, no disambig symbol.
    # Else output #1, or #2, #3, ... if the same token-seq
    # has already been assigned a disambig symbol.
    wordid2tokenid = defaultdict(list)
    id2token = lexicon.id2token.copy()

    first_allowed_disambig = lexicon.num_disambigs
    first_allowed_disambig_id = last_used_disambig_id + 1
    max_disambig = first_allowed_disambig - 1
    last_used_disambig_id_of = defaultdict(int)

    for word_id, token_ids in lexicon:
        token_key = tuple(token_ids)
        assert len(token_key) > 0
        if issubseq[token_key] == 0 and count[token_key] == 1 or tokenword_mode and word_id >= first_tokenword_id:
            wordid2tokenid[word_id].append(token_ids)
            continue

        cur_disambig_id = last_used_disambig_id_of[token_key]
        if cur_disambig_id == 0:
            cur_disambig = first_allowed_disambig
            cur_disambig_id = first_allowed_disambig_id
        else:
            cur_disambig = int(id2token[cur_disambig_id].name.lstrip("#")) + 1

        if cur_disambig > max_disambig:
            max_disambig = cur_disambig
            cur_disambig_id = max(id2token.keys()) + 1
            id2token[cur_disambig_id] = LexiconUnit(f"#{max_disambig}", "disambig_subsequence")
        last_used_disambig_id_of[token_key] = cur_disambig_id
        wordid2tokenid[word_id].append(token_ids + [cur_disambig_id])
    return Lexicon(wordid2tokenid, lexicon.id2word, id2token)


def make_lexicon_fst_no_silence(
    lexicon: Lexicon,
    attach_symbol_table: bool = True,
) -> 'kaldifst.StdVectorFst':
    """
    Compiles a Lexicon into a lexicon WFST (L.fst).

    See also make_lexicon_fst.py from kaldi.

    Args:
      lexicon:
        Lexicon object.

    Returns:
      Kaldi-type lexicon WFST.
    """
    _kaldifst_maybe_raise()

    backoff_disambig = "#0"
    tokenword_disambig = "#1"
    tokenword_mode = tokenword_disambig in lexicon.word2id
    if tokenword_mode:
        first_tokenword_id = lexicon.word2id[tokenword_disambig] + 1

    fst = kaldifst.StdVectorFst()
    start_state = fst.add_state()
    fst.start = start_state
    fst.set_final(state=start_state, weight=0)
    fst.add_arc(
        state=start_state,
        arc=kaldifst.StdArc(
            ilabel=lexicon.token2id[backoff_disambig],
            olabel=lexicon.word2id[backoff_disambig],
            weight=0,
            nextstate=start_state,
        ),
    )
    if tokenword_mode:
        tokenword_state_begin = fst.add_state()
        fst.add_arc(
            state=start_state,
            arc=kaldifst.StdArc(
                ilabel=lexicon.token2id[tokenword_disambig],
                olabel=lexicon.word2id[tokenword_disambig],
                weight=0,
                nextstate=tokenword_state_begin,
            ),
        )

    for word_id, token_ids in lexicon:
        cur_state = start_state

        if not tokenword_mode or word_id < first_tokenword_id - 1:
            for i, token_id in enumerate(token_ids[:-1]):
                next_state = fst.add_state()
                fst.add_arc(
                    state=cur_state,
                    arc=kaldifst.StdArc(
                        ilabel=token_id,
                        olabel=word_id if i == 0 else 0,
                        weight=0,
                        nextstate=next_state,
                    ),
                )
                cur_state = next_state
            i = len(token_ids) - 1  # note: i == -1 if tokens is empty.
            fst.add_arc(
                state=cur_state,
                arc=kaldifst.StdArc(
                    ilabel=token_ids[-1] if i >= 0 else 0,
                    olabel=word_id if i <= 0 else 0,
                    weight=0,
                    nextstate=start_state,
                ),
            )
    if tokenword_mode:
        tokenword_begin, tokenword_other = [], []
        for word_id in range(first_tokenword_id, max(lexicon.id2word) + 1):
            token_id = lexicon.token2id[lexicon.id2word[word_id].name.rstrip(f"{TW_BREAK}{tokenword_disambig}")]
            token_unit = lexicon.id2token[token_id]
            if token_unit.mark.startswith("begin"):
                tokenword_begin.append((token_id, word_id))
            elif token_unit.mark == "":
                tokenword_other.append((token_id, word_id))
            else:
                raise RuntimeError(f"Unexpected mark `{token_unit.mark}` for tokenword `{token_unit.name}`")

        tokenword_state_main = fst.add_state()
        for token_id, word_id in tokenword_begin:
            fst.add_arc(
                state=tokenword_state_begin,
                arc=kaldifst.StdArc(
                    ilabel=token_id,
                    olabel=word_id,
                    weight=0,
                    nextstate=tokenword_state_main,
                ),
            )
        tokenword_state_end = fst.add_state()
        for token_id, word_id in tokenword_other:
            fst.add_arc(
                state=tokenword_state_main,
                arc=kaldifst.StdArc(
                    ilabel=token_id,
                    olabel=word_id,
                    weight=0,
                    nextstate=tokenword_state_main,
                ),
            )
            fst.add_arc(
                state=tokenword_state_main,
                arc=kaldifst.StdArc(
                    ilabel=token_id,
                    olabel=word_id,
                    weight=0,
                    nextstate=tokenword_state_end,
                ),
            )
        fst.add_arc(
            state=tokenword_state_end,
            arc=kaldifst.StdArc(
                ilabel=lexicon.token2id[tokenword_disambig],
                olabel=lexicon.word2id[tokenword_disambig],
                weight=0,
                nextstate=start_state,
            ),
        )

    if attach_symbol_table:
        isym = kaldifst.SymbolTable()
        for p, i in lexicon.token2id.items():
            isym.add_symbol(symbol=p, key=i)
        fst.input_symbols = isym

        osym = kaldifst.SymbolTable()
        for w, i in lexicon.word2id.items():
            osym.add_symbol(symbol=w, key=i)
        fst.output_symbols = osym

    kaldifst.arcsort(fst, sort_type="ilabel")
    return fst


def build_topo(
    name: str, token2id: Dict[str, int], with_self_loops: bool = True, attach_symbol_table: bool = True
) -> 'kaldifst.StdVectorFst':
    """Helper function to build a topology WFST (T.fst).

    Args:
      name:
        Topology name. Choices: default, compact, minimal

      token2id:
        Token index.
        Mapping from token_str to token_id.

      with_self_loops:
        Whether to add token-to-epsilon self-loops to the topology.

      attach_symbol_table:
        Whether to attach the token names for indices of the returned WFST.

    Returns:
      Kaldi-type topology WFST.
    """
    _kaldifst_maybe_raise()

    if name == "default":
        fst = build_default_topo(token2id, with_self_loops)
    elif name == "compact":
        fst = build_compact_topo(token2id, with_self_loops)
    elif name == "minimal":
        fst = build_minimal_topo(token2id)
    else:
        raise ValueError(f"Unknown topo name: {name}")

    if attach_symbol_table:
        isym = kaldifst.SymbolTable()
        for t, i in token2id.items():
            isym.add_symbol(symbol=t, key=i)
        fst.input_symbols = isym
        fst.output_symbols = fst.input_symbols.copy()
    return fst


def build_default_topo(token2id: Dict[str, int], with_self_loops: bool = True) -> 'kaldifst.StdVectorFst':
    """Build the default (correct) CTC topology."""
    _kaldifst_maybe_raise()

    disambig_pattern = re.compile(r"^#\d+$")
    blank_id = token2id["<blk>"]
    fst = kaldifst.StdVectorFst()
    start_state = fst.add_state()
    fst.start = start_state
    fst.set_final(state=start_state, weight=0)
    fst.add_arc(
        state=start_state,
        arc=kaldifst.StdArc(
            ilabel=blank_id,
            olabel=0,
            weight=0,
            nextstate=start_state,  # token2id["<eps>"] is always 0
        ),
    )

    disambig_ids = []
    token_ids = {}
    for s, i in token2id.items():
        if s == "<eps>" or s == "<blk>":
            continue
        elif disambig_pattern.match(s):
            disambig_ids.append(i)
        else:
            state = fst.add_state()
            fst.set_final(state=state, weight=0)
            token_ids[state] = i
            fst.add_arc(
                state=start_state,
                arc=kaldifst.StdArc(
                    ilabel=i,
                    olabel=i,
                    weight=0,
                    nextstate=state,
                ),
            )
            if with_self_loops:
                fst.add_arc(
                    state=state,
                    arc=kaldifst.StdArc(
                        ilabel=i,
                        olabel=0,
                        weight=0,
                        nextstate=state,  # token2id["<eps>"] is always 0
                    ),
                )
            fst.add_arc(
                state=state,
                arc=kaldifst.StdArc(
                    ilabel=blank_id,
                    olabel=0,
                    weight=0,
                    nextstate=start_state,  # token2id["<eps>"] is always 0
                ),
            )

    for istate in kaldifst.StateIterator(fst):
        if istate > 0:
            for ostate in kaldifst.StateIterator(fst):
                if ostate > 0 and istate != ostate:
                    label = token_ids[ostate]
                    fst.add_arc(
                        state=istate,
                        arc=kaldifst.StdArc(
                            ilabel=label,
                            olabel=label,
                            weight=0,
                            nextstate=ostate,
                        ),
                    )
        for disambig_id in disambig_ids:
            fst.add_arc(
                state=istate,
                arc=kaldifst.StdArc(
                    ilabel=0,
                    olabel=disambig_id,
                    weight=0,
                    nextstate=istate,  # token2id["<eps>"] is always 0
                ),
            )

    return fst


def build_compact_topo(token2id: Dict[str, int], with_self_loops: bool = True) -> 'kaldifst.StdVectorFst':
    """Build the Compact CTC topology."""
    _kaldifst_maybe_raise()

    disambig_pattern = re.compile(r"^#\d+$")
    blank_id = token2id["<blk>"]
    fst = kaldifst.StdVectorFst()
    start_state = fst.add_state()
    fst.start = start_state
    fst.set_final(state=start_state, weight=0)
    fst.add_arc(
        state=start_state,
        arc=kaldifst.StdArc(
            ilabel=blank_id,
            olabel=0,
            weight=0,
            nextstate=start_state,  # token2id["<eps>"] is always 0
        ),
    )

    for s, i in token2id.items():
        if s == "<eps>" or s == "<blk>":
            continue
        elif disambig_pattern.match(s):
            fst.add_arc(
                state=start_state,
                arc=kaldifst.StdArc(
                    ilabel=0,
                    olabel=i,
                    weight=0,
                    nextstate=start_state,  # token2id["<eps>"] is always 0
                ),
            )
        else:
            state = fst.add_state()
            fst.add_arc(
                state=start_state,
                arc=kaldifst.StdArc(
                    ilabel=i,
                    olabel=i,
                    weight=0,
                    nextstate=state,
                ),
            )
            if with_self_loops:
                fst.add_arc(
                    state=state,
                    arc=kaldifst.StdArc(
                        ilabel=i,
                        olabel=0,
                        weight=0,
                        nextstate=state,  # token2id["<eps>"] is always 0
                    ),
                )
            fst.add_arc(
                state=state,
                arc=kaldifst.StdArc(
                    ilabel=0,  # token2id["<eps>"] is always 0
                    olabel=0,  # token2id["<eps>"] is always 0
                    weight=0,
                    nextstate=start_state,
                ),
            )

    return fst


def build_minimal_topo(token2id: Dict[str, int]) -> 'kaldifst.StdVectorFst':
    """Build the Minimal CTC topology."""
    _kaldifst_maybe_raise()

    disambig_pattern = re.compile(r"^#\d+$")
    blank_id = token2id["<blk>"]
    fst = kaldifst.StdVectorFst()
    start_state = fst.add_state()
    fst.start = start_state
    fst.set_final(state=start_state, weight=0)
    fst.add_arc(
        state=start_state,
        arc=kaldifst.StdArc(
            ilabel=blank_id,
            olabel=0,
            weight=0,
            nextstate=start_state,  # token2id["<eps>"] is always 0
        ),
    )

    for s, i in token2id.items():
        if s == "<eps>" or s == "<blk>":
            continue
        elif disambig_pattern.match(s):
            fst.add_arc(
                state=start_state,
                arc=kaldifst.StdArc(
                    ilabel=0,
                    olabel=i,
                    weight=0,
                    nextstate=start_state,  # token2id["<eps>"] is always 0
                ),
            )
        else:
            fst.add_arc(
                state=start_state,
                arc=kaldifst.StdArc(
                    ilabel=i,
                    olabel=i,
                    weight=0,
                    nextstate=start_state,
                ),
            )

    return fst


def mkgraph_ctc_ov(
    tokenizer: 'TokenizerSpec',
    lm_path: Union[Path, str],
    topology_name: str = "default",
    write_tlg_path: Optional[Union[Path, str]] = None,
    open_vocabulary: bool = False,
    open_vocabulary_weights: Tuple[float, float] = (2.0, 4.0),
    target: str = "kaldi",  # "kaldi", "k2"
) -> Tuple[Union['kaldifst.StdVectorFst', 'k2.Fsa'], int]:
    """
    Builds a decoding WFST (TLG.fst or TLG.pt).

    See also mkgraph.sh from kaldi.

    Args:
      tokenizer:
        NeMo SentencePiece tokenizer.

      lm_path:
        Path to the ARPA LM file.

      topology_name:
        Topology name. Choices: default, compact, minimal.

      write_tlg_path:
        Where to buffer the TLG.

      open_vocabulary:
        Whether to build a decoding WFST suitable for the open vocabulary decoding.

      open_vocabulary_weights:
        Pair of weights (oov_word_weight, token_unigram_weight).

      target:
        What type to build the WFST for. Choices: kaldi, k2.

    Returns:
      A pair of kaldi- or k2-type decoding WFST and its id of the tokenword disambiguation token.
    """
    _kaldifst_maybe_raise()

    logging.info("Compiling G.fst ...")
    G = arpa2fst(lm_path)
    if open_vocabulary:
        # in-place for g_fst
        tokenword_disambig_id = add_tokenwords_(
            g_fst=G,
            tokens=tokenizer.tokenizer.get_vocab().keys(),
            word_weight=open_vocabulary_weights[0],
            token_unigram_weight=open_vocabulary_weights[1],
        )
    else:
        tokenword_disambig_id = -1

    logging.info("Building L.fst ...")
    id2word = {int(line.split("\t")[1]): line.split("\t")[0] for line in str(G.output_symbols).strip().split("\n")}
    lexicon = generate_lexicon_sentencepiece(
        tokenizer.tokenizer, id2word, add_epsilon=True, first_tokenword_id=tokenword_disambig_id
    )
    lexicon_disambig = add_disambig_symbols(lexicon)

    L = make_lexicon_fst_no_silence(lexicon_disambig)
    kaldifst.arcsort(L, sort_type="olabel")

    logging.info("Building LG.fst ...")
    LG = kaldifst.compose(L, G)
    kaldifst.determinize_star(LG)
    kaldifst.minimize_encoded(LG)
    kaldifst.arcsort(LG, sort_type="ilabel")

    logging.info("Building TLG.fst ...")
    T = build_topo(topology_name, lexicon_disambig.token2id)
    kaldifst.arcsort(T, sort_type="olabel")
    TLG = kaldifst.compose(T, LG)

    if target == "kaldi":
        if write_tlg_path:
            logging.info(f"Buffering TLG.fst into {write_tlg_path} ...")
            TLG.write(write_tlg_path)
    elif target == "k2":
        logging.info("Converting TLG.fst to k2 ...")
        import torch

        from nemo.core.utils.k2_guard import k2

        blank_id = [i for i, t in lexicon_disambig.id2token.items() if t.mark == "blank"][0]
        first_token_disambig_id = [i for i, t in lexicon_disambig.id2token.items() if t.mark == "disambig_backoff"][0]
        word_disambig_id = lexicon_disambig.word2id[lexicon_disambig.id2token[first_token_disambig_id].name]
        assert lexicon_disambig.id2word[word_disambig_id].mark == "disambig_backoff"
        input_symbols = "\n".join(
            [f"{k} {v - 1}" for k, v in lexicon_disambig.token2id.items() if 0 < v < first_token_disambig_id]
        )
        output_symbols = str(TLG.output_symbols)
        TLG.input_symbols = None
        TLG.output_symbols = None
        # k2 does not support torch.inference_mode enabled
        with torch.inference_mode(False):
            TLG = k2.Fsa.from_openfst(TLG.to_str(show_weight_one=True), acceptor=False)
            TLG.labels[TLG.labels >= first_token_disambig_id] = blank_id
            TLG.aux_labels[TLG.aux_labels.values == word_disambig_id] = 0
            TLG.__dict__["_properties"] = None
            TLG = k2.arc_sort(k2.connect(k2.remove_epsilon(TLG)))
            TLG.labels[TLG.labels > 0] = TLG.labels[TLG.labels > 0] - 1
            TLG.__dict__["_properties"] = None
            TLG.labels_sym = k2.SymbolTable.from_str(input_symbols)
            TLG.aux_labels_sym = k2.SymbolTable.from_str(output_symbols)
            TLG = k2.arc_sort(TLG)
            if write_tlg_path:
                logging.info(f"Buffering TLG.pt into {write_tlg_path} ...")
                torch.save(TLG.as_dict(), write_tlg_path)
    else:
        raise ValueError(f"Unsupported target: `{target}`")

    return TLG, tokenword_disambig_id


class KaldiFstMask(Enum):
    Acceptor = 65536
    Error = 4
    TopSorted = 274877906944
    Acyclic = 34359738368
    IlabelSorted = 268435456
    OlabelSorted = 1073741824
    IlabelDeterministic = 262144
    OlabelDeterministic = 1048576
    HasEpsilons = 4194304
    HasIEpsilons = 16777216
    Accessible = 1099511627776
    Coaccessible = 4398046511104
    Weighted = 4294967296


class LatticeProperties(NamedTuple):
    Acceptor: bool
    Valid: bool
    Nonempty: bool
    TopSorted: bool
    Acyclic: bool
    ArcSorted: bool
    Deterministic: bool
    EpsilonFree: bool
    InputEpsilonFree: bool
    Connected: bool
    Weighted: bool


class AbstractLattice(ABC):
    """A lattice wrapper with high-level capabilities."""

    def __init__(self, lattice: Any):
        self._lattice = lattice
        self._properties = None

    @abstractmethod
    def as_tensor(self) -> 'torch.Tensor':
        """Represents the lattice as a tensor.

        Returns:
          torch.Tensor
        """
        pass

    @abstractmethod
    def draw(
        self, filename: Optional[Union[Path, str]] = None, title: Optional[Union[Path, str]] = None, zoom: float = 1.0
    ) -> Union['graphviz.Digraph', 'IPython.display.HTML']:
        """Render FSA as an image via graphviz, and return the Digraph object; and optionally save to file filename.
        filename must have a suffix that graphviz understands, such as pdf, svg or png.

        Note:
          You need to install graphviz to use this function::

            ./scripts/installers/install_graphviz.sh

        Args:
          filename:
            Filename to (optionally) save to, e.g. ‘foo.png’, ‘foo.svg’, ‘foo.png’.

          title:
            Title to be displayed in image, e.g. ‘A simple lattice example’.

          zoom:
            Zoom-in lattice in IPython notebook (needed for large lattices).

        Returns:
          graphviz.Digraph or IPython.display.HTML
        """
        pass

    @abstractmethod
    def edit_distance(self, reference_sequence: List[int]) -> int:
        """Get the edit distance from a reference sequence to the lattice.

        Args:
          reference_sequence:
            List of word- or token-ids.

        Returns:
          Number of edits.
        """

    @property
    def lattice(self):
        self._properties = None
        return self._lattice

    @abstractproperty
    def properties(self) -> LatticeProperties:
        pass

    @abstractproperty
    def symbol_table(self) -> Optional[Dict[int, str]]:
        pass

    @abstractproperty
    def auxiliary_tables(self) -> Optional[Tuple[Any]]:
        pass


class KaldiWordLattice(AbstractLattice):
    """A Kaldi lattice wrapper with high-level capabilities."""

    def __init__(
        self,
        lattice: 'kaldifst.Lattice',
        symbol_table: Optional[Dict[int, str]] = None,
        auxiliary_tables: Optional[Dict[str, Any]] = None,
    ):
        _kaldifst_maybe_raise()

        if not isinstance(lattice, kaldifst.Lattice):
            raise ValueError(f"Wrong lattice type: `{type(lattice)}`")
        super().__init__(lattice)

        kaldi_symbols2dict = lambda symbols: {
            int(line.split("\t")[1]): line.split("\t")[0] for line in str(symbols).strip().split("\n")
        }
        self._symbol_table = None
        # most likely lattice will have empty input_symbols
        if symbol_table is not None:
            self._symbol_table = symbol_table
        elif self._lattice.output_symbols is not None:
            # we suppose that lattice.input_symbols will not be changed
            self._symbol_table = kaldi_symbols2dict(self._lattice.output_symbols)

        self._auxiliary_tables = None
        if auxiliary_tables is not None:
            attributes, values = list(auxiliary_tables.keys()), list(auxiliary_tables.values())
            if "input_symbols" not in attributes and self._lattice.input_symbols is not None:
                # rare but possible case
                attributes.append("input_symbols")
                values.append(kaldi_symbols2dict(self._lattice.input_symbols))
            self._auxiliary_tables = namedtuple("KaldiAuxiliaryTables", attributes)(*values)
        elif self._lattice.input_symbols is not None:
            self._auxiliary_tables = namedtuple("KaldiAuxiliaryTables", "input_symbols")(
                kaldi_symbols2dict(self._lattice.input_symbols)
            )

    @property
    def properties(self) -> LatticeProperties:
        if self._properties is None:
            acceptor = self._lattice.properties(KaldiFstMask.Acceptor.value, True) == KaldiFstMask.Acceptor.value
            valid = self._lattice.properties(KaldiFstMask.Error.value, True) != KaldiFstMask.Error.value
            nonempty = self._lattice.num_states > 0
            top_sorted = self._lattice.properties(KaldiFstMask.TopSorted.value, True) == KaldiFstMask.TopSorted.value
            acyclic = self._lattice.properties(KaldiFstMask.Acyclic.value, True) == KaldiFstMask.Acyclic.value
            arc_sorted = (
                self._lattice.properties(KaldiFstMask.IlabelSorted.value, True) == KaldiFstMask.IlabelSorted.value
                and self._lattice.properties(KaldiFstMask.OlabelSorted.value, True) == KaldiFstMask.OlabelSorted.value
            )
            deterministic = (
                self._lattice.properties(KaldiFstMask.IlabelDeterministic.value, True)
                == KaldiFstMask.IlabelDeterministic.value
                and self._lattice.properties(KaldiFstMask.OlabelDeterministic.value, True)
                == KaldiFstMask.OlabelDeterministic.value
            )
            epsilon_free = (
                self._lattice.properties(KaldiFstMask.HasEpsilons.value, True) != KaldiFstMask.HasEpsilons.value
            )
            input_epsilon_free = (
                self._lattice.properties(KaldiFstMask.HasIEpsilons.value, True) != KaldiFstMask.HasIEpsilons.value
            )
            connected = (
                self._lattice.properties(KaldiFstMask.Accessible.value, True) == KaldiFstMask.Accessible.value
                and self._lattice.properties(KaldiFstMask.Coaccessible.value, True) == KaldiFstMask.Coaccessible.value
            )
            weighted = self._lattice.properties(KaldiFstMask.Weighted.value, True) == KaldiFstMask.Weighted.value
            self._properties = LatticeProperties(
                Acceptor=acceptor,
                Valid=valid,
                Nonempty=nonempty,
                TopSorted=top_sorted,
                Acyclic=acyclic,
                ArcSorted=arc_sorted,
                Deterministic=deterministic,
                EpsilonFree=epsilon_free,
                InputEpsilonFree=input_epsilon_free,
                Connected=connected,
                Weighted=weighted,
            )
        return self._properties

    @property
    def symbol_table(self) -> Optional[Dict[int, str]]:
        return self._symbol_table

    @property
    def auxiliary_tables(self) -> Optional[Tuple[Any]]:
        return self._auxiliary_tables

    def as_tensor(self) -> 'torch.Tensor':
        """Represents the lattice as a tensor.

        Returns:
          torch.Tensor
        """
        raise NotImplementedError("Tensor representation is not supported yet.")

    def edit_distance(self, reference_sequence: List[int]) -> int:
        """Get the edit distance from a reference sequence to the lattice.

        Args:
          reference_sequence:
            List of word- or token-ids.

        Returns:
          Number of edits.
        """
        _kaldifst_maybe_raise()

        if not self.properties.InputEpsilonFree:
            logging.warning(f"Lattice contains input epsilons. Edit distance calculations may not be accurate.")
        if not all(reference_sequence):
            raise ValueError(f"reference_sequence contains zeros, which is not allowed.")
        ref = levenshtein_graph_kaldi(kaldifst.make_linear_acceptor(reference_sequence))
        hyp = levenshtein_graph_kaldi(self._lattice)
        kaldifst.invert(hyp)
        ali_fst = kaldifst.compose(hyp, ref)
        succeeded, _, _, total_weight = kaldifst.get_linear_symbol_sequence(kaldifst.shortest_path(ali_fst))
        if not succeeded:
            raise RuntimeError("Something went wrong while calculating edit_distance. Please check input manually.")
        return round(total_weight.value)

    def draw(
        self, filename: Optional[Union[Path, str]] = None, title: Optional[Union[Path, str]] = None, zoom: float = 1.0
    ) -> Union['graphviz.Digraph', 'IPython.display.HTML']:
        """Render FSA as an image via graphviz, and return the Digraph object; and optionally save to file filename.
        filename must have a suffix that graphviz understands, such as pdf, svg or png.

        Note:
          You need to install graphviz to use this function::

            ./scripts/installers/install_graphviz.sh

        Args:
          filename:
            Filename to (optionally) save to, e.g. ‘foo.png’, ‘foo.svg’, ‘foo.png’.

          title:
            Title to be displayed in image, e.g. ‘A simple lattice example’.

          zoom:
            Zoom-in lattice in IPython notebook (needed for large lattices).

        Returns:
          graphviz.Digraph or IPython.display.HTML
        """
        _kaldifst_maybe_raise()
        _graphviz_maybe_raise()

        isym, osym = None, None
        if self._symbol_table:
            osym = kaldifst.SymbolTable()
            for i, w in self._symbol_table.items():
                osym.add_symbol(symbol=w, key=i)

        if (
            self._auxiliary_tables
            and hasattr(self._auxiliary_tables, "input_symbols")
            and self._auxiliary_tables.input_symbols
        ):
            isym = kaldifst.SymbolTable()
            for i, t in self._auxiliary_tables.input_symbols.items():
                isym.add_symbol(symbol=t, key=i)

        fst_dot = kaldifst.draw(
            self._lattice, acceptor=False, portrait=True, isymbols=isym, osymbols=osym, show_weight_one=True
        )
        source = graphviz.Source(fst_dot)
        source_lines = str(source).splitlines()
        # Remove 'digraph tree {'
        source_lines.pop(0)
        # Remove the closing brackets '}'
        source_lines.pop(-1)
        graph_attr = {
            'rankdir': 'LR',
            'size': '8.5,11',
            'center': '1',
            'orientation': 'Portrait',
            'ranksep': '0.4',
            'nodesep': '0.25',
            'margin': '0.0',
        }
        if title is not None:
            graph_attr['label'] = title
        digraph = graphviz.Digraph(graph_attr=graph_attr)
        digraph.body += source_lines
        if filename:
            _, extension = os.path.splitext(filename)
            if extension == '' or extension[0] != '.':
                raise ValueError(f"Filename needs to have a suffix like .png, .pdf, .svg, or .gv: `{filename}`")
            with tempfile.TemporaryDirectory() as tmp_dir:
                temp_fn = digraph.render(filename='temp', directory=tmp_dir, format=extension[1:], cleanup=True)

                shutil.move(temp_fn, filename)
        if _is_notebook():
            import warnings

            from IPython.display import HTML

            with tempfile.TemporaryDirectory() as tmp_dir:
                temp_fn = digraph.render(filename='temp', directory=tmp_dir, format="svg", cleanup=True)
                svg, (width, height) = _svg_srcdoc_resize(temp_fn, zoom)
            # IFrame requires src file to be present when rendering
            # so we use HTML with iframe srcdoc instead
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return HTML(
                    f"""<iframe srcdoc='{svg}' width="100%" height="{round(height * zoom) * 2}px" frameborder="0" allowfullscreen></iframe>"""
                )
        return digraph


def _is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell' or 'Shell':
            return True  # Jupyter notebook, Google Colab notebook, or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type
    except NameError:
        return False  # Probably standard Python interpreter


def _svg_srcdoc_resize(filename: Union[Path, str], zoom: float) -> Tuple[str, Tuple[int, int]]:
    with open(filename, "rt", encoding="utf-8") as f:
        line = f.readline()
        while not line.startswith("<svg"):
            line = f.readline()
        width, height = re.findall(r'\d+', line)
        width, height = int(width), int(height)
        return f'<svg width="{round(width * zoom)}pt" height="{round(height * zoom)}pt"\n' + f.read(), (width, height)


def levenshtein_graph_kaldi(
    fst: Union['kaldifst.StdFst', 'kaldifst.Lattice'], ins_del_score: float = 0.501
) -> 'kaldifst.StdFst':
    """Construct the levenshtein graph from a kaldi-type WFST or a lattice.

    See also levenshtein_graph from k2.

    Args:
      fst:
        Kaldi-type source WFST or lattice.

      ins_del_score:
        Insertion and deletion penalty.
        Should be more than 0.5 for substitutions to be preferred over insertions/deletions, or less otherwise.

    Returns:
      Kaldi-type levenshtein WFST.
    """
    _kaldifst_maybe_raise()

    if fst.properties(KaldiFstMask.Acceptor.value, True) != KaldiFstMask.Acceptor.value:
        logging.warning(
            "Levenshtein graph construction is not safe for WFSTs with different input and output symbols."
        )
    if fst.properties(KaldiFstMask.Acyclic.value, True) != KaldiFstMask.Acyclic.value:
        raise ValueError("Levenshtein graph is not defined for WFSTs with cycles.")
    if isinstance(fst, kaldifst.StdFst):
        lfst = fst.copy(safe=True)
    elif isinstance(fst, kaldifst.Lattice):
        # dropping lattice weights
        lfst = kaldifst.compile(re.sub(r"[-\d.]+,[-\d.]+", "0", fst.to_str(show_weight_one=True)))
    else:
        raise ValueError(f"Levenshtein graph building is not supported for the type `{type(fst)}`.")
    sub_score = 0.5
    eps = 0
    for state in kaldifst.StateIterator(lfst):
        # epsilon self-loop for insertions and deletions
        arcs_to_add = [
            kaldifst.StdArc(
                ilabel=eps,
                olabel=eps,
                weight=ins_del_score,
                nextstate=state,
            )
        ]
        for arc in kaldifst.ArcIterator(lfst, state):
            # epsilon-to-ilabel arc for substitutions
            arcs_to_add.append(
                kaldifst.StdArc(
                    ilabel=eps,
                    olabel=arc.ilabel,
                    weight=sub_score,
                    nextstate=arc.nextstate,
                )
            )
            # zero weight for correct ids (redundant for lattices)
            arc.weight = 0.0
        for arc in arcs_to_add:
            lfst.add_arc(state=state, arc=arc)
    kaldifst.arcsort(lfst)
    return lfst


def load_word_lattice(
    lat_filename: Union[Path, str], id2word: Optional[Dict[int, str]] = None, id2token: Optional[Dict[int, str]] = None
) -> Dict[str, KaldiWordLattice]:
    """Helper function to load riva-decoder recognition lattices.

    Args:
      lat_filename:
        Path to the riva-decoder recognition lattice file.

      id2word:
        Word index.
        Mapping from word_id to word_str.

      id2token:
        Token index.
        Mapping from token_id to token_str.

    Returns:
      Dictionary with lattice names and corresponding lattices in KaldiWordLattice format.
    """
    _kaldifst_maybe_raise()

    lattice_dict = {}
    lattice = None
    max_state = 0
    token_seq_list = []
    with open(lat_filename, "rt") as f:
        for line in f.readlines():
            line_items = line.strip().split()
            line_len = len(line_items)
            if line_len == 0:  # end of lattice
                token_seq_list = []
                lattice = None
                max_state = 0
            elif line_len == 1:  # lattice identifier
                assert lattice is None
                assert max_state == 0
                assert len(token_seq_list) == 0
                lat_id = line_items[0]
                lattice = kaldifst.Lattice()
                lattice_dict[lat_id] = KaldiWordLattice(
                    lattice=lattice,
                    symbol_table=id2word,
                    auxiliary_tables={"token_seq_list": token_seq_list, "input_symbols": id2token},
                )
                start = lattice.add_state()
                lattice.start = start
                max_state += 1
            elif line_len in (3, 4):  # arc
                if line_len == 4:  # regular arc
                    state, next_state, label = [int(i) for i in line_items[:-1]]
                    trunk = line_items[-1].split(',')
                    graph_cost, acoustic_cost = [float(i) for i in trunk[:-1]]
                else:  # arc without weight
                    logging.warning(
                        f"""An arc without weight is detected for lattice `{lat_id}`.
                                    Weights and token sequences will be set trivially."""
                    )
                    state, next_state, label = [int(i) for i in line_items]
                    trunk = [""]
                    graph_cost, acoustic_cost = 0.0, 0.0
                if next_state >= max_state:
                    for i in range(max_state, next_state + 1):
                        lattice.add_state()
                    max_state = next_state + 1
                ark = kaldifst.LatticeArc(
                    ilabel=label,
                    olabel=label,
                    weight=kaldifst.LatticeWeight(graph_cost=graph_cost, acoustic_cost=acoustic_cost),
                    nextstate=next_state,
                )
                lattice.add_arc(state=state, arc=ark)
                token_seq_list.append((ark, [int(i) for i in trunk[-1].split(TW_BREAK)] if trunk[-1] != "" else []))
            elif line_len == 2:  # final state
                state = int(line_items[0])
                trunk = line_items[-1].split(',')
                graph_cost, acoustic_cost = [float(i) for i in trunk[:-1]]
                lattice.set_final(
                    state=state, weight=kaldifst.LatticeWeight(graph_cost=graph_cost, acoustic_cost=acoustic_cost)
                )
            else:
                raise RuntimeError(f"Broken line: `{line}`")
    return lattice_dict



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from nemo.lightning import io
from nemo.lightning.pytorch.callbacks.peft import ADAPTER_META_FILENAME
from nemo.lightning.resume import AdapterPath, AutoResume
from nemo.utils import logging
from nemo.utils.app_state import AppState


@dataclass(kw_only=True)
class SpeechLMAutoResume(AutoResume):
    """
    Wrapper for AutoResume to get rid of requirement on restore_config.
    """

    adapter_path: Optional[str] = None

    def _maybe_get_adapter_path(self, checkpoint: Path) -> Optional[Path]:
        if self.adapter_path:
            logging.info(f"Using provided adapter path: {self.adapter_path}")
            return AdapterPath(Path(self.adapter_path), base_model_path=checkpoint)
        else:
            adapter_meta_path = checkpoint / ADAPTER_META_FILENAME
            if adapter_meta_path.exists():
                logging.info(f"Found adapter meta file at {adapter_meta_path}")
                return AdapterPath(checkpoint, base_model_path=checkpoint)
            else:
                return Path(checkpoint)

    def get_trainer_ckpt_path(self, model: Optional[io.ConnectorMixin] = None) -> Optional[Path]:
        if self.resume_from_path:
            logging.info(f"Attempting to resume from {self.resume_from_path}")
            maybe_weights_path = self.get_weights_path(self.resume_from_path)
            if not maybe_weights_path.is_dir():
                raise ValueError(
                    f"Provided path {self.resume_from_path} doesn't seem to contain a {self.WEIGHTS_PATH} dir."
                )
            return self._maybe_get_adapter_path(maybe_weights_path)

        checkpoint = None
        app_state = AppState()
        app_state.restore = self.resume_if_exists
        if self.resume_if_exists:
            checkpoint = self._find_trainer_ckpt_path()

        if checkpoint:
            maybe_weights_path = self.get_weights_path(checkpoint)
            if maybe_weights_path.is_dir():
                checkpoint = maybe_weights_path

        if checkpoint:
            adapter_meta_path = checkpoint / ADAPTER_META_FILENAME
            if adapter_meta_path.exists():
                return AdapterPath(checkpoint, base_model_path=checkpoint)
            else:
                return Path(checkpoint)

        return None



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import io
import signal
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Optional, Set, Tuple, Union, cast

import lightning.pytorch as pl
import torch
from lightning.fabric.plugins import ClusterEnvironment
from lightning.pytorch.callbacks import TQDMProgressBar

try:
    from megatron.core import parallel_state
    from megatron.core.dist_checkpointing.mapping import ShardedBase, ShardedObject, ShardedTensor
    from megatron.core.dist_checkpointing.strategies.torch import sharded_tensor_to_torch_sharded_tensor
    from megatron.core.transformer.utils import _get_extra_state_offsets

    HAVE_MEGATRON_CORE = True

except (ImportError, ModuleNotFoundError):

    ShardedObject = object
    ShardedBase = object
    ShardedTensor = object
    HAVE_MEGATRON_CORE = False


from torch import Tensor, nn
from torch.distributed._sharded_tensor import ShardedTensor as TorchShardedTensor
from torch.distributed._tensor import DTensor, Replicate, Shard
from torch.distributed.device_mesh import DeviceMesh, _mesh_resources
from torch.distributed.tensor.parallel import ColwiseParallel, RowwiseParallel, SequenceParallel, parallelize_module

from nemo.lightning import _strategy_lib
from nemo.lightning.pytorch.callbacks import MegatronProgressBar, ProgressPrinter
from nemo.utils.callbacks.dist_ckpt_io import AsyncFinalizableCheckpointIO
from nemo.utils.import_utils import safe_import_from

MixedPrecisionPolicy, HAS_MIXED_PRECISION_POLICY = safe_import_from(
    "torch.distributed.fsdp", "MixedPrecisionPolicy", fallback_module="torch.distributed._composable.fsdp"
)
fully_shard, HAS_FULLY_SHARD = safe_import_from(
    "torch.distributed.fsdp", "fully_shard", fallback_module="torch.distributed._composable.fsdp"
)
CPUOffloadPolicy, HAS_CPU_OFFLOAD_POLICY = safe_import_from(
    "torch.distributed.fsdp", "CPUOffloadPolicy", fallback_module="torch.distributed._composable.fsdp"
)


@dataclass(kw_only=True)
class RestoreConfig:
    """
    Configuration for restoring model state from a checkpoint.

    Attributes:
        path (str): Path to the checkpoint directory.
        load_model_state (bool): Whether to load model weights.
        load_optim_state (bool): Whether to load optimizer state.
        load_artifacts (bool): Whether to load additional artifacts (e.g., tokenizer).
    """

    path: str
    load_model_state: bool = True
    load_optim_state: bool = False
    # eg tokenizer, etc.
    load_artifacts: bool = True


def setup_parallel_ranks(strategy: pl.strategies.Strategy):
    """
    Sets up parallel ranks for distributed training.

    Args:
        strategy (pl.strategies.Strategy): The Lightning strategy being used for training.
    """
    from megatron.core.model_parallel_config import ModelParallelConfig

    env = cast(ClusterEnvironment, strategy.cluster_environment)
    parallelism = getattr(strategy, "parallelism", ModelParallelConfig())
    _strategy_lib.init_parallel_ranks(env.world_size(), env.global_rank(), env.local_rank(), parallelism)


def init_model_parallel(pl_module: pl.LightningModule):
    """
    Initializes model parallelism for distributed training.

    Args:
        pl_module (pl.LightningModule): The PyTorch Lightning module.
    """
    from megatron.core import parallel_state

    from nemo.utils import AppState

    if not parallel_state.model_parallel_is_initialized():
        app_state = AppState()

        if app_state.model_parallel_size is not None:
            _strategy_lib.init_model_parallel(pl_module)


def setup_data_sampler(trainer: pl.Trainer):
    """
    Configures the data sampler for distributed training.

    Args:
        trainer (pl.Trainer): The PyTorch Lightning trainer instance.
    """
    datamodule = getattr(trainer, "datamodule", None)
    if datamodule is not None:
        if hasattr(trainer.strategy, "data_sampler") and trainer.strategy.data_sampler is not None:
            datamodule.data_sampler = trainer.strategy.data_sampler
        elif hasattr(datamodule, "data_sampler"):
            trainer.strategy.data_sampler = datamodule.data_sampler

    if trainer.strategy.data_sampler is not None:
        trainer.strategy.data_sampler.setup(trainer.strategy.cluster_environment.global_rank())
        trainer.strategy.data_sampler.connect(trainer)

    if hasattr(datamodule, "reconfigure_limit_batches"):
        datamodule.reconfigure_limit_batches()


def fix_progress_bar(trainer: pl.Trainer, replace_progress_bar: bool = True, progress_interval: int = 1) -> None:
    """
    Fixes or replaces the progress bar callback in the PyTorch Lightning trainer.

    Args:
        trainer (pl.Trainer): The PyTorch Lightning trainer instance.
        replace_progress_bar (bool): Whether to replace the default progress bar.
        progress_interval (int): Interval at which to log progress.
    """
    callbacks: List[pl.Callback] = cast(List[pl.Callback], getattr(trainer, "callbacks"))
    contains_megatron_progress, contains_progress = False, False
    for callback in callbacks:
        if isinstance(callback, MegatronProgressBar):
            contains_megatron_progress = True
        if callback.__class__ == TQDMProgressBar:
            contains_progress = True
    if not contains_megatron_progress and contains_progress:
        for i, callback in enumerate(callbacks):
            if isinstance(callback, TQDMProgressBar):
                if replace_progress_bar:
                    printer = ProgressPrinter(log_interval=progress_interval)
                    printer._trainer = trainer
                    if not trainer.is_global_zero:
                        printer.disable()
                    callbacks[i] = printer
                else:
                    callback.__class__ = MegatronProgressBar
                break


def ckpt_to_dir(filepath: Union[str, Path]) -> Path:
    """PTL considers checkpoints as .ckpt files.
    This method removes the extension and returns a path
    to be used as a directory for distributed checkpoints.

    Converts a checkpoint file path to a directory path by removing the `.ckpt` extension.

    Args:
        filepath (Union[str, Path]): The checkpoint file path.

    Returns:
        Path: The directory path where the checkpoint will be stored.

    """
    filepath = Path(filepath)

    if filepath.suffix == ".ckpt":
        return filepath.with_name(filepath.stem)

    return filepath


def create_checkpoint_io(wrapping_ckpt_io=None, **kwargs):
    """
    Creates a checkpoint IO handler for saving/loading checkpoints.

    Args:
        wrapping_ckpt_io: An optional wrapper for checkpoint IO.
        **kwargs: Additional arguments to configure checkpoint IO.

    Returns:
        Checkpoint IO handler instance.
    """
    if kwargs.get("model_library", None) == "huggingface":
        from nemo.lightning.io.hf import HFCheckpointIO

        checkpoint_io = HFCheckpointIO(adapter_only=kwargs.get("lora", False))
    else:
        from nemo.lightning.io.pl import MegatronCheckpointIO

        checkpoint_io = MegatronCheckpointIO(**kwargs)

    if wrapping_ckpt_io:
        checkpoint_io = wrapping_ckpt_io(checkpoint_io)
    if kwargs.get("async_save", False):
        checkpoint_io = AsyncFinalizableCheckpointIO(checkpoint_io)

    return checkpoint_io


def mcore_to_pyt_sharded_state_dict(
    checkpoint: Dict[str, List[torch.Tensor]],
    sharded_state_dict: Dict[str, Union[List[ShardedTensor], ShardedObject]],
    dtensor: bool = False,
    device_mesh: DeviceMesh = None,
) -> Dict[str, Union[TorchShardedTensor, io.BytesIO]]:
    """
    Converts a Megatron-Core sharded state dictionary into a PyTorch-compatible format.

    Args:
        checkpoint (Dict[str, List[torch.Tensor]]):
            The Megatron-Core checkpoint containing a list of tensors for each key.
        sharded_state_dict (Dict[str, Union[List[ShardedTensor], ShardedObject]]):
            The corresponding PyTorch sharded state dictionary.
        dtensor (bool, optional):
            Whether to use DTensor for the conversion. Defaults to False.
        device_mesh (DeviceMesh, optional):
            The device mesh configuration for distributed tensors.

    Returns:
        Dict[str, Union[TorchShardedTensor, io.BytesIO]]:
            A PyTorch-compatible state dictionary with properly formatted sharded tensors.
    """

    def _mcore_to_pyt_dtensor(
        tens: List[torch.Tensor],
        sh_tens: List[ShardedTensor],
        device_mesh: DeviceMesh,
    ) -> DTensor:
        """Converts a Megatron-Core tensor into a PyTorch DTensor."""
        assert len(tens) == 1 and len(sh_tens) == 1

        dten = DTensor.from_local(
            tens[0],
            device_mesh,
            (
                Replicate(),
                Shard(dim=0),
            ),  # hardcoded for HSDP
        )
        return dten

    def _mcore_to_pyt_sharded_tensor(tens: List[torch.Tensor], sh_tens: List[ShardedTensor]) -> TorchShardedTensor:
        """Converts a Megatron-Core tensor into a PyTorch sharded tensor."""
        for ten, sh_ten in zip(tens, sh_tens):
            # remove prepend axes and put in loaded tensor
            sh_ten.global_shape = sh_ten.global_shape[sh_ten.prepend_axis_num :]
            sh_ten.global_offset = sh_ten.global_offset[sh_ten.prepend_axis_num :]
            sh_ten.axis_fragmentations = sh_ten.axis_fragmentations[sh_ten.prepend_axis_num :]
            sh_ten.prepend_axis_num = 0
            sh_ten.data = ten
            sh_ten.validate_metadata_integrity()

        return sharded_tensor_to_torch_sharded_tensor(sh_tens)

    def _convert(checkpoint, sharded_state_dict, k, device_mesh=None):
        """Recursively converts checkpoint tensors into PyTorch-compatible formats."""
        assert k in sharded_state_dict, f"{k} not in sharded_state_dict"

        if isinstance(sharded_state_dict[k], Dict):
            for kk in sharded_state_dict[k]:
                _convert(checkpoint[k], sharded_state_dict[k], kk, device_mesh=device_mesh)
        elif isinstance(sharded_state_dict[k], ShardedObject):
            """Do nothing. checkpoint[k] contains loaded io.BytesIO already."""
        elif isinstance(sharded_state_dict[k], List):  # list of ShardedTensor
            if dtensor:
                checkpoint[k] = _mcore_to_pyt_dtensor(checkpoint[k], sharded_state_dict[k], device_mesh=device_mesh)
            else:
                checkpoint[k] = _mcore_to_pyt_sharded_tensor(checkpoint[k], sharded_state_dict[k])

    for k in checkpoint:
        _convert(checkpoint, sharded_state_dict, k, device_mesh=device_mesh)

    return checkpoint


def pyt_to_mcore_state_dict(
    state_dict: Dict[str, Any], prefix: str = "", device_mesh: DeviceMesh = None
) -> Dict[str, List[ShardedBase]]:
    """
    Converts a PyTorch state dictionary into a Megatron-Core compatible format.

    Args:
        state_dict (Dict[str, Any]):
            The PyTorch state dictionary.
        prefix (str, optional):
            A prefix to prepend to all keys. Defaults to "".
        device_mesh (DeviceMesh, optional):
            The device mesh configuration for distributed tensors.

    Returns:
        Dict[str, List[ShardedBase]]:
            A Megatron-Core formatted state dictionary with properly sharded tensors.
    """

    def _dtensor_to_mcore_sharded_tensor(
        key: str,
        dten: DTensor,
        prepend_offsets: Iterable[Tuple[int, int, int]] = (),
        prefix: str = "",
        allow_shape_mismatch: bool = False,
        device_mesh: DeviceMesh = None,
    ) -> List[ShardedTensor]:
        prepend_axis_num = len(prepend_offsets)

        assert device_mesh is not None
        assert isinstance(dten, DTensor), dten

        ten = dten.to_local()
        global_shape = dten.shape

        rank_offsets = []
        replica_id = 0
        axis = list(range(len(global_shape)))
        axis_fragm = [global_shape[i] // ten.shape[i] for i in axis]

        for i, placement in enumerate(dten.placements):
            if isinstance(placement, Shard):
                ax = placement.dim
                rank_offsets.append((ax + prepend_axis_num, dten.device_mesh.get_local_rank(i), axis_fragm[ax]))
            elif placement.is_replicate():
                replica_id = device_mesh.get_local_rank(i)

        local_shard = ShardedTensor.from_rank_offsets(
            f"{prefix}{key}",
            ten,
            *prepend_offsets,  # prepend layer shards
            *rank_offsets,
            replica_id=replica_id,
            prepend_axis_num=prepend_axis_num,
            allow_shape_mismatch=allow_shape_mismatch,
        )
        return [local_shard]

    def _torch_to_mcore_sharded_tensor(
        key: str,
        sh_ten: TorchShardedTensor,
        prepend_offsets: Iterable[Tuple[int, int, int]] = (),
        prefix: str = "",
        allow_shape_mismatch: bool = False,
    ) -> List[ShardedTensor]:
        prepend_axis_num = len(prepend_offsets)

        assert isinstance(sh_ten, TorchShardedTensor), sh_ten
        sharded_meta = sh_ten.metadata()
        local_shards = sh_ten.local_shards()

        # DEBUG
        assert all([ls.metadata.placement == local_shards[0].metadata.placement for ls in local_shards]), [
            ls.meta.placement for ls in local_shards
        ]

        global_shape = sharded_meta.size

        axis = list(range(len(global_shape)))
        axis_fragm = [global_shape[i] // local_shards[0].metadata.shard_sizes[i] for i in axis]
        rank_offsets = []

        for i in range(len(local_shards)):
            local_shard = local_shards[i]
            ten, meta = local_shard.tensor, local_shard.metadata

            for j in range(len(axis)):
                axis_rank_offset = meta.shard_offsets[j] // meta.shard_sizes[j]
                rank_offsets.append((axis[j] + prepend_axis_num, axis_rank_offset, axis_fragm[j]))

            local_shards[i] = ShardedTensor.from_rank_offsets(
                f"{prefix}{key}",
                ten,
                *prepend_offsets,  # prepend layer shards
                *rank_offsets,
                replica_id=0,
                prepend_axis_num=prepend_axis_num,
                allow_shape_mismatch=allow_shape_mismatch,
            )

        return local_shards

    def _torch_to_mcore_sharded_object(
        key: str,
        obj: io.BytesIO,
        sharded_offsets: Iterable[Tuple[int, int, int]] = (),
        prefix: str = "",
    ) -> ShardedObject:
        """mcore helper"""
        replica_id = (
            0,
            0,
            parallel_state.get_data_parallel_rank(with_context_parallel=True),
        )

        return ShardedObject(f"{prefix}{key}", obj, *_get_extra_state_offsets(sharded_offsets), replica_id)

    def _convert(state_dict, k, sh_key, v, prepend_offsets, prefix="", allow_shape_mismatch=False, device_mesh=None):
        """mcore helper"""
        if isinstance(v, Dict):
            for kk, vv in v.items():
                _convert(
                    v,
                    kk,
                    sh_key,
                    vv,
                    prepend_offsets,
                    prefix=f"{prefix}{kk}.",
                    allow_shape_mismatch=allow_shape_mismatch,
                    device_mesh=device_mesh,
                )
        elif isinstance(v, DTensor):
            state_dict[k] = _dtensor_to_mcore_sharded_tensor(
                sh_key,
                v,
                prepend_offsets,
                prefix=prefix,
                allow_shape_mismatch=allow_shape_mismatch,
                device_mesh=device_mesh,
            )
        elif isinstance(v, TorchShardedTensor):
            state_dict[k] = _torch_to_mcore_sharded_tensor(
                sh_key, v, prepend_offsets, prefix=prefix, allow_shape_mismatch=allow_shape_mismatch
            )
        elif isinstance(v, io.BytesIO):
            state_dict[k] = _torch_to_mcore_sharded_object(sh_key, v, prepend_offsets, prefix)

    num_layers = 0
    for k in state_dict:
        if k.startswith("module.decoder.layers."):
            num_layers = max(num_layers, int(k.split(".")[3]) + 1)

    for k, v in state_dict.items():
        prepend_offsets = []
        sh_key = k
        allow_shape_mismatch = k.endswith(".word_embeddings.weight")  # vocab size can be different
        if k.startswith("module.decoder.layers."):
            sh_key = k.split(".")
            global_layer_offset = int(sh_key.pop(3))
            sh_key = ".".join(sh_key)
            prepend_offsets.append((0, global_layer_offset, num_layers))

        _convert(state_dict, k, sh_key, v, prepend_offsets, prefix, allow_shape_mismatch, device_mesh)

    return state_dict


# Taken and modified from torchtitan
# https://github.com/pytorch/torchtitan/blob/main/torchtitan/parallelisms/parallelize_llama.py
def fsdp2_strategy_parallelize(
    model,
    device_mesh: DeviceMesh = None,
    mp_policy: MixedPrecisionPolicy = None,
    use_hf_tp_plan: bool = False,
    tp_shard_plan: Optional[Dict[str, Union[RowwiseParallel, ColwiseParallel, SequenceParallel]]] = None,
    offload_policy: 'CPUOffloadPolicy' = None,
):
    """Apply parallelisms and activation checkpointing to the model.

    Args:
        model: The model to be parallelized.
        device_mesh (DeviceMesh): The device mesh for distributed training.
        mp_policy (MixedPrecisionPolicy): Mixed precision policy for model parallelism.
        tp_shard_plan (Optional[Dict[str, Union[RowwiseParallel, ColwiseParallel, SequenceParallel]]]):
            A tensor parallel sharding plan. The keys should be the module names and the values should be the
            corresponding parallel styles (e.g., RowwiseParallel, ColwiseParallel, SequenceParallel).
        offload_policy (CPUOffloadPolicy): The offload policy for FSDP. If None, it will use the default policy.

    NOTE: The passed-in model preferably should be on meta device. Otherwise,
    the model must fit on GPU or CPU memory.
    NOTE: Currently, the user is required to manually handle precision settings such as the `mp_policy` here
    because the model parallel strategy does not respect all settings of `Fabric(precision=...)` at the moment.
    NOTE: Currently, the user should make sure that custom_tp_plan is compatible with the model architecture.
    """

    if not mp_policy:
        assert HAS_MIXED_PRECISION_POLICY is not None, "Expected to have MixedPrecisionPolicy"
        mp_policy = MixedPrecisionPolicy(param_dtype=torch.bfloat16, reduce_dtype=torch.float32)

    def parallelize_helper(module, mesh, mp_policy):
        if isinstance(module, nn.ModuleList):
            for layer_id, transformer_block in enumerate(module):
                # Apply activation checkpointing
                # transformer_block = checkpoint_wrapper(transformer_block)
                # As an optimization, do not reshard after forward for the last
                # transformer block since FSDP would prefetch it immediately
                reshard_after_forward = int(layer_id) < len(module) - 1
                fully_shard(
                    transformer_block,
                    mesh=mesh,
                    mp_policy=mp_policy,
                    reshard_after_forward=reshard_after_forward,
                    offload_policy=offload_policy,
                )
                module[layer_id] = transformer_block
        else:
            for name, sub_module in module.named_children():
                parallelize_helper(sub_module, mesh, mp_policy)

    # Set FSDP sharding mesh to context parallel mesh if CP > 1, else default to the data parallel mesh.
    dp_mesh = device_mesh[
        ("dp_cp" if "dp_cp" in _mesh_resources.root_to_flatten_mapping.get(device_mesh, {}) else "data_parallel")
    ]
    tp_mesh = device_mesh["tensor_parallel"]

    if dp_mesh.size() > 1:
        assert dp_mesh.ndim == 1, "Hybrid-sharding not supported"

    # TP sharding
    if tp_mesh.size() > 1:
        if tp_shard_plan is None and use_hf_tp_plan:
            tp_shard_plan = get_hf_tp_shard_plan(model)
        parallelize_module(model, tp_mesh, tp_shard_plan)

    # FSDP sharding
    assert dp_mesh.ndim == 1, "Hybrid-sharding not supported"
    assert HAS_FULLY_SHARD is not None, "Expected to have fully_shard"

    # Find transformer layers and apply parallelisms
    parallelize_helper(model, dp_mesh, mp_policy)

    # reshard_after_forward=True based on
    # https://github.com/pytorch/torchtitan/blob/main/torchtitan/parallelisms/parallelize_llama.py#L359
    model = fully_shard(
        model, mesh=dp_mesh, mp_policy=mp_policy, reshard_after_forward=True, offload_policy=offload_policy
    )

    return model


def get_hf_tp_shard_plan(model):
    """
    Get the tensor parallel sharding plan from the model.
    """
    hf_tp_shard_plan = {}
    if hasattr(model, '_tp_plan') and model._tp_plan is not None:
        hf_tp_shard_plan.update(model._tp_plan)
    if hasattr(model.model, '_tp_plan') and model.model._tp_plan is not None:
        hf_tp_shard_plan.update({f"model.{k}": v for k, v in model.model._tp_plan.items()})

    hf_tp_shard_plan = {k: translate_to_torch_parallel_style(v) for k, v in hf_tp_shard_plan.items()}
    return hf_tp_shard_plan


@lru_cache
def translate_to_torch_parallel_style(style: str):
    """
    In model configurations, we use a neutral type (string) to specify parallel
    styles, here we translate them into torch.distributed tensor-parallel
    types.
    """
    if not isinstance(style, str):
        raise ValueError(f"Unsupported parallel style type {type(style)}, expected str")

    if style == "colwise":
        return ColwiseParallel()
    elif style == "rowwise":
        return RowwiseParallel()
    elif style == "colwise_rep":
        return ColwiseParallel(output_layouts=Replicate())
    elif style == "rowwise_rep":
        return RowwiseParallel(input_layouts=Replicate())
    elif style == "sequence_parallel":
        return SequenceParallel()
    else:
        raise ValueError(f"Unsupported parallel style value: {style}")


def to_cpu(v):
    """
    Move a tensor or distributed tensor to the CPU.

    This function takes an input tensor, which can be either a `DTensor` (distributed tensor)
    or a standard `Tensor`, and ensures that it is moved to the CPU.

    Args:
        v (DTensor | Tensor | any): The input value, which can be a `DTensor`, `Tensor`, or
                                    any other object. If `DTensor`, it checks the device and
                                    moves the tensor accordingly.

    Returns:
        Tensor | any: The corresponding CPU tensor if `v` is a `DTensor` or `Tensor`,
                    otherwise returns `v` unchanged.

    Raises:
        ValueError: If `v` is a `DTensor` but its device is neither 'cuda' nor 'cpu'.

    Example:
        >>> t = torch.tensor([1, 2, 3], device='cuda')
        >>> to_cpu(t)  # Moves tensor to CPU
        tensor([1, 2, 3])

        >>> dt = DTensor(torch.tensor([4, 5, 6], device='cuda'))
        >>> to_cpu(dt)  # Moves DTensor to CPU
        tensor([4, 5, 6])
    """
    if isinstance(v, DTensor):
        if v.device.type == "cuda":
            return v.full_tensor().cpu()
        elif v.device.type == "cpu":
            return v._local_tensor
        else:
            raise ValueError("Unknown device " + str(v.device))
    elif isinstance(v, Tensor):
        return v.cpu()
    else:
        return v


def _destroy_dist_connection() -> None:
    """Destroy process group."""
    # Don't allow Ctrl+C to interrupt this handler
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        torch.distributed.destroy_process_group()
    signal.signal(signal.SIGINT, signal.SIG_DFL)


# based on https://github.com/pytorch/torchtitan/blob/main/torchtitan/distributed/utils.py#L113
def create_context_parallel_ctx(
    cp_mesh: DeviceMesh,
    cp_buffers: List[torch.Tensor],
    cp_seq_dims: List[int],
    cp_no_restore_buffers: Set[torch.Tensor],
    cp_rotate_method: str,
):
    """
    Create a context parallel context.

    Args:
        cp_mesh (DeviceMesh): The device mesh for context parallel.
        cp_buffers (List[torch.Tensor]): The buffers for context parallel.
        cp_seq_dims (List[int]): The sequence dimensions for context parallel.
        cp_no_restore_buffers (Set[torch.Tensor]): The no restore buffers for context parallel.
        cp_rotate_method (str): The rotation method for context parallel, such as "allgather" or "addtoall".
    """
    from torch.distributed.tensor.experimental import context_parallel

    # TODO: uncomment this when torch.distributed.tensor.experimental._attention.set_rotate_method is available
    # from torch.distributed.tensor.experimental._attention import set_rotate_method
    # set_rotate_method(cp_rotate_method)
    return context_parallel(
        cp_mesh,
        buffers=cp_buffers,
        buffer_seq_dims=cp_seq_dims,
        no_restore_buffers=cp_no_restore_buffers,
    )


# based on https://github.com/pytorch/torchtitan/blob/main/torchtitan/distributed/utils.py#L138
def get_train_context(enable_loss_parallel: bool, enable_compiled_autograd: bool):
    """
    Create a train context.

    Args:
        enable_loss_parallel (bool): Whether to enable loss parallelism.
        enable_compiled_autograd (bool): Whether to enable compiled autograd.
    """

    @contextlib.contextmanager
    def context(cp_context: Optional[Generator[None, None, None]] = None):
        with contextlib.ExitStack() as stack:
            if enable_loss_parallel:
                stack.enter_context(torch.distributed.tensor.parallel.loss_parallel())

            if enable_compiled_autograd:
                stack.enter_context(torch._dynamo.utils.maybe_enable_compiled_autograd(True))

            if cp_context is not None:
                from torch.nn.attention import SDPBackend, sdpa_kernel

                # currently we only support these two SDP backends.
                # TODO (xilunwu): support cuDNN backend
                stack.enter_context(sdpa_kernel([SDPBackend.FLASH_ATTENTION, SDPBackend.EFFICIENT_ATTENTION]))
                stack.enter_context(cp_context)

            yield

    return context



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from pathlib import Path

import pytest
import torch
from torch.nn.utils.rnn import pad_sequence
from tqdm.auto import tqdm

from nemo.collections.asr.parts.submodules.ngram_lm import KenLMBatchedWrapper, NGramGPULanguageModel
from nemo.core.utils.optional_libs import KENLM_AVAILABLE, TRITON_AVAILABLE

DEVICES = [torch.device("cpu")]

if torch.cuda.is_available():
    DEVICES.append('cuda')


@pytest.fixture(scope="module")
def n_gpu_lm(test_data_dir):
    kenlm_model_path = Path(test_data_dir) / "asr/kenlm_ngram_lm/parakeet-tdt_ctc-110m-libri-1024.kenlm.tmp.arpa"
    vocab_size = 1024
    return NGramGPULanguageModel.from_arpa(kenlm_model_path, vocab_size=vocab_size, normalize_unk=False)


@pytest.fixture(scope="module")
def kenlm_wrapper(test_data_dir):
    kenlm_model_path = Path(test_data_dir) / "asr/kenlm_ngram_lm/parakeet-tdt_ctc-110m-libri-1024.kenlm.tmp.arpa"
    vocab_size = 1024
    return KenLMBatchedWrapper.from_file(lm_path=kenlm_model_path, vocab_size=vocab_size)


class TestNGramGPULanguageModel:
    @pytest.mark.with_downloads
    @pytest.mark.unit
    def test_load(self, test_data_dir):
        kenlm_model_path = Path(test_data_dir) / "asr/kenlm_ngram_lm/parakeet-tdt_ctc-110m-libri-1024.kenlm.tmp.arpa"
        _ = NGramGPULanguageModel.from_arpa(kenlm_model_path, vocab_size=1024)

    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.skipif(not KENLM_AVAILABLE, reason="KenLM is not available")
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("batch_size", [1, 3])
    @pytest.mark.parametrize("bos", [True, False])
    def test_initial_states(
        self,
        n_gpu_lm: NGramGPULanguageModel,
        kenlm_wrapper: KenLMBatchedWrapper,
        bos: bool,
        batch_size: int,
        device: torch.device,
    ):
        n_gpu_lm = n_gpu_lm.to(device)
        init_states = n_gpu_lm.get_init_states(batch_size=batch_size, bos=bos)
        init_states_kenlm = kenlm_wrapper.get_init_states(batch_size=batch_size, bos=bos)
        scores_lm, _ = n_gpu_lm.advance(init_states)
        scores_ref, _ = kenlm_wrapper.advance(init_states_kenlm)
        assert torch.allclose(scores_lm, scores_ref.to(device))

    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.skipif(not TRITON_AVAILABLE, reason="Triton is not available")
    @pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available")
    def test_triton_vs_pytorch_random_states(self, n_gpu_lm: NGramGPULanguageModel, batch_size=2, num_iterations=100):
        """Randomly initializes the states and compares the scores from Triton and PyTorch implementations."""
        torch.manual_seed(777)
        device = torch.device("cuda")
        n_gpu_lm = n_gpu_lm.to(device)
        for _ in tqdm(range(num_iterations)):
            start_state = random.randint(0, n_gpu_lm.num_states - 1)
            with torch.no_grad():
                scores1, states1 = n_gpu_lm._advance_pytorch(
                    states=torch.full([batch_size], fill_value=start_state, device=device, dtype=torch.int64)
                )
                scores2, states2 = n_gpu_lm._advance_triton(
                    states=torch.full([batch_size], fill_value=start_state, device=device, dtype=torch.int64)
                )
            assert (states1 == states2).all()
            assert torch.allclose(scores1, scores2)

    @pytest.mark.unit
    @pytest.mark.skipif(not KENLM_AVAILABLE, reason="KenLM is not available")
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("bos", [True, False])
    def test_final(
        self, n_gpu_lm: NGramGPULanguageModel, kenlm_wrapper: KenLMBatchedWrapper, bos: bool, device: torch.device
    ):
        """Test final (eos) scores"""
        n_gpu_lm = n_gpu_lm.to(device)
        sentences = [
            [25, 70, 12],
            [58, 41, 186, 293, 306, 999, 163, 264, 689, 683, 999],
            [],  # empty sentence
        ]
        last_states = []
        for sentence in sentences:
            state = kenlm_wrapper.get_init_state(bos=bos)
            for label in sentence:
                _, state = kenlm_wrapper.advance_single(state=state, label=label)
            last_states.append(state)
        final_ref = kenlm_wrapper.get_final(states=last_states).to(device=device)

        last_states = []
        for sentence in sentences:
            states = n_gpu_lm.get_init_states(batch_size=1, bos=bos)
            for label in sentence:
                _, states = n_gpu_lm.advance(states=states)
                states = states[0, label].unsqueeze(0)
            last_states.append(states)
        final_lm = n_gpu_lm.get_final(states=torch.cat(last_states, dim=0))

        assert torch.allclose(final_lm, final_ref), "Final scores do not match"

    @pytest.mark.unit
    @pytest.mark.skipif(not KENLM_AVAILABLE, reason="KenLM is not available")
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("bos", [True, False])
    @pytest.mark.parametrize("eos", [True, False])
    def test_sentences(
        self,
        n_gpu_lm: NGramGPULanguageModel,
        kenlm_wrapper: KenLMBatchedWrapper,
        bos: bool,
        eos: bool,
        device: torch.device,
    ):
        n_gpu_lm = n_gpu_lm.to(device)
        sentences = [
            [25, 70, 12],
            [58, 41, 186, 293, 306, 999, 163, 264, 689, 683, 999],
            [],  # empty sentence
        ]
        # non-batched
        for sentence in sentences:
            scores_ref = kenlm_wrapper.score_sentences([sentence], bos=bos, eos=eos).to(device)
            scores_lm = n_gpu_lm(
                labels=torch.LongTensor([sentence]).to(device),
                bos=bos,
                eos=eos,
            )
            assert torch.allclose(scores_ref, scores_lm), "Non-batched scores do not match"

        # batched
        scores_ref = kenlm_wrapper.score_sentences(sentences, bos=bos, eos=eos).to(device)
        scores_lm = n_gpu_lm(
            labels=pad_sequence([torch.LongTensor(sentence) for sentence in sentences], batch_first=True).to(device),
            labels_lengths=torch.LongTensor([len(sentence) for sentence in sentences]).to(device),
            bos=bos,
            eos=eos,
        )
        assert torch.allclose(scores_lm, scores_ref), "Batched scores do not match"

    @pytest.mark.unit
    def test_save_load_nemo(self, tmp_path, test_data_dir):
        vocab_size = 1024
        kenlm_model_path = Path(test_data_dir) / "asr/kenlm_ngram_lm/parakeet-tdt_ctc-110m-libri-1024.kenlm.tmp.arpa"
        n_gpu_lm = NGramGPULanguageModel.from_arpa(kenlm_model_path, vocab_size=vocab_size, normalize_unk=False)
        nemo_path = tmp_path / "ngram_lm.nemo"
        n_gpu_lm.save_to(f"{nemo_path}")
        n_gpu_lm_loaded = NGramGPULanguageModel.from_nemo(f"{nemo_path}", vocab_size=vocab_size)

        # arcs data
        assert torch.allclose(n_gpu_lm_loaded.arcs_weights, n_gpu_lm.arcs_weights)
        assert (n_gpu_lm_loaded.from_states == n_gpu_lm.from_states).all()
        assert (n_gpu_lm_loaded.to_states == n_gpu_lm.to_states).all()
        assert (n_gpu_lm_loaded.ilabels == n_gpu_lm.ilabels).all()

        # states data
        assert (n_gpu_lm_loaded.start_end_arcs == n_gpu_lm.start_end_arcs).all()
        assert (n_gpu_lm_loaded.state_order == n_gpu_lm.state_order).all()
        assert (n_gpu_lm_loaded.backoff_to_states == n_gpu_lm.backoff_to_states).all()
        assert torch.allclose(n_gpu_lm_loaded.backoff_weights, n_gpu_lm.backoff_weights)
        assert torch.allclose(n_gpu_lm_loaded.final_weights, n_gpu_lm.final_weights)

    @pytest.mark.unit
    def test_save_load_from_file(self, tmp_path, test_data_dir):
        vocab_size = 1024
        kenlm_model_path = Path(test_data_dir) / "asr/kenlm_ngram_lm/parakeet-tdt_ctc-110m-libri-1024.kenlm.tmp.arpa"
        n_gpu_lm = NGramGPULanguageModel.from_file(kenlm_model_path, vocab_size=vocab_size, normalize_unk=False)
        nemo_path = tmp_path / "ngram_lm.nemo"
        n_gpu_lm.save_to(f"{nemo_path}")
        n_gpu_lm_loaded = NGramGPULanguageModel.from_file(f"{nemo_path}", vocab_size=vocab_size)

        # arcs data
        assert torch.allclose(n_gpu_lm_loaded.arcs_weights, n_gpu_lm.arcs_weights)
        assert (n_gpu_lm_loaded.from_states == n_gpu_lm.from_states).all()
        assert (n_gpu_lm_loaded.to_states == n_gpu_lm.to_states).all()
        assert (n_gpu_lm_loaded.ilabels == n_gpu_lm.ilabels).all()

        # states data
        assert (n_gpu_lm_loaded.start_end_arcs == n_gpu_lm.start_end_arcs).all()
        assert (n_gpu_lm_loaded.state_order == n_gpu_lm.state_order).all()
        assert (n_gpu_lm_loaded.backoff_to_states == n_gpu_lm.backoff_to_states).all()
        assert torch.allclose(n_gpu_lm_loaded.backoff_weights, n_gpu_lm.backoff_weights)
        assert torch.allclose(n_gpu_lm_loaded.final_weights, n_gpu_lm.final_weights)



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from abc import abstractmethod
from typing import Dict, List, Optional, Set, Tuple

import pytorch_lightning as pl
import torch
from megatron.core import InferenceParams, parallel_state

from nemo.collections.common.tokenizers.chat_template_mixin import explode_chat_template_input, is_chat_input
from nemo.collections.multimodal.speech_llm.parts.utils.data_utils import shift_tokens_by_multi_audios

# the text representation of eos_id, it applies for all tokenizers
END_OF_SEQ = '<|endoftext|>'


def pad_batch(batch, pad_id, max_len):
    """pad batch seq to the same length"""
    context_lengths = []
    max_context_length = max([len(tokens) for tokens in batch])
    for tokens in batch:
        context_length = len(tokens)
        if context_length < max_context_length + max_len:
            tokens.extend([pad_id] * (max_context_length + max_len - context_length))
        context_lengths.append(context_length)
    return batch, context_lengths


def switch(val1, val2, boolean):
    """switch between two values based on the boolean tensor"""
    boolean = boolean.type_as(val1)
    boolean = boolean.unsqueeze(0).unsqueeze(-1)
    return (1 - boolean) * val1 + boolean * val2


class TextGenerationStrategy:
    """
    Base class for TextGeneration Strategy
    """

    def __init__(self, model):
        self.model = model
        self.tokenizer = getattr(model, "tokenizer", None)
        self.inference_params = None
        if self.model.training:
            # TODO in the future this should raise an exception
            warnings.warn(
                "Generation started while the model is in training mode, switching to eval mode "
                "(this situation may raise an exception in future versions, please call `eval()` before generation)"
            )
            self.model.eval()
        self._end_of_generation_cache = None

    def forward_step(self, batch: Dict[str, torch.Tensor]) -> torch.Tensor:
        """
        given input, run the forward step of the language model to get output tensor
        """
        if "labels" in batch:
            batch.pop("labels")

        output_tensor = self.model.language_model(**batch)
        # preds = torch.argmax(output_tensor, dim=-1)

        if self.inference_params:
            # if last stage, then (final) output is [b, s, h], otherwise it's [s, b, h]
            if parallel_state.is_pipeline_last_stage():
                self.inference_params.sequence_len_offset += output_tensor.size(1)
            else:
                self.inference_params.sequence_len_offset += output_tensor.size(0)

        return output_tensor

    def tokenize_batch(self, sentences, max_len, add_BOS):
        """
        convert the sentences into lists of tokens, pad them to the same length, add bos tokens if it is needed
        Args:
            sentences (List[str]): list of input sentences in str format.
            max_len (int): max number of tokens to generate.
            add_BOS (bool): whether to add the BOS token at the beginning
        Returns:
            Tuple[torch.Tensor], the tokenized and padded torch tensor and the token context length tensor.
        """
        tokenizer = self.model.tokenizer
        if is_chat_input(sentences):
            assert getattr(
                tokenizer, 'has_chat_template', False
            ), "Got chat-template input but tokenizer does not support chat template formating."
            context_tokens = list(map(tokenizer.text_to_ids, explode_chat_template_input(sentences)))
        elif add_BOS:
            context_tokens = [[tokenizer.bos_id] + tokenizer.text_to_ids(s) for s in sentences]
        elif hasattr(tokenizer.tokenizer, "get_prefix_tokens"):
            # chatglm: add tokenizer.gmask_id, tokenizer.sop_id
            context_tokens = [tokenizer.tokenizer.get_prefix_tokens() + tokenizer.text_to_ids(s) for s in sentences]
        else:
            context_tokens = [tokenizer.text_to_ids(s) for s in sentences]
        context_tokens, context_lengths = pad_batch(context_tokens, tokenizer.eos_id, max_len)
        context_tokens_tensor = torch.cuda.LongTensor(context_tokens)
        context_length_tensor = torch.cuda.LongTensor(context_lengths)
        return context_tokens_tensor, context_length_tensor

    @classmethod
    @abstractmethod
    def clip_max_len(self, maxlen: int) -> int:
        """clip the max len based on the LM model max sequence length
        Args:
            maxlen (int): the max len computed from the context and number of tokens to generate
        returns (int):
            the clip the max length based of the LM model max sequence length
        """
        pass

    @classmethod
    @abstractmethod
    def init_batch(self, context_tokens: torch.Tensor, context_length: int, compute_attention_mask: bool):
        """initialize the batch data before the inference steps.
           It will save the intermediate results as object attributes
           context_length (int): the context token length
           compute_attention_mask: bool: set to True to compute attention mask (not needed for FA)
        Args:
            context_tokens (torch.Tensor):  The padded context tokens including the space for tokens to be generated
        """
        pass

    @classmethod
    @abstractmethod
    def prepare_batch_at_step(
        self, tokens: torch.Tensor, maxlen: int, micro_batch_size: int, step: int, context_length: int
    ) -> Dict[str, torch.Tensor]:
        """
        generate the batch used in inference for each of the steps
        Args:
            tokens  (torch.Tensor): the context tokens
            maxlen (int): the maximum length in the context tokens
            micro_batch_size (int): text generation batch size
            step (int): the inference step count
            context_length (int): the new token position in the tokens
        returns:
            a dict of tensors used in the inference step
        """
        pass

    @classmethod
    @abstractmethod
    def post_process(self, tokens: torch.Tensor, new_tokens: torch.Tensor, context_length: int):
        """
        At the end of the single step inference, post process the inference results
        Args:
            tokens  (torch.Tensor): the context tokens
            new_token (torch.Tensor): sampled new token id
            context_length (int): the new token position in the tokens
        """
        pass

    def end_of_generation_condition(
        self, tokens: torch.Tensor, prev: torch.Tensor, eod_id: int, end_strings: List[str]
    ) -> torch.Tensor:
        """
        return whether the generation should stop based on the previous token
        Args:
            tokens (torch.Tensor): the generated tokens so far
            prev  (torch.Tensor): the previous token
            eod_id (int): the end of document token id
            end_strings (List[str]): the list of end of generation strings
        returns:
            a boolean tensor indicating whether the generation should stop
        """
        if (len(end_strings) == 1 and end_strings[0] == END_OF_SEQ) or not end_strings:
            # Simple scenario: only finish on end of document token.
            return prev == eod_id

        end_tokens, end_strings_to_check = self._get_end_of_generation_tokens_and_strings(eod_id, end_strings)
        assert end_tokens

        is_end = torch.isin(prev, torch.tensor(list(end_tokens), dtype=prev.dtype, device=prev.device))

        if end_strings_to_check:
            # The loop below is inefficient (see warning in `_get_end_of_generation_tokens_and_strings()`)
            # TODO In addition, we will not stop if the model generates an end string followed by extra characters,
            # e.g., if `end_string` is "Done" and there exists a "Done!" token it could generate tokens
            #       [..., ".", "Done!"]
            # which would fail the `endswith("Done")` check. However, stopping when "Done!" is generated would not
            # work either, since we would need to post-process the generated string to truncate the extra "!".
            # ==> this is left for future work if there is a compelling use case requiring this feature.
            for idx, token_seq in enumerate(tokens):
                text = self.model.tokenizer.ids_to_text(token_seq.tolist())
                is_end[idx] |= any(text.endswith(end_string) for end_string in end_strings_to_check)

        return is_end

    def post_generation_process(self, output):
        """
        At the end of the text generation, post process the results
        Args:
            output  (dict): the text generation output dictionary
        """
        return output

    def _get_end_of_generation_tokens_and_strings(
        self, eod_id: int, end_strings: List[str]
    ) -> Tuple[Set[int], List[str]]:
        """
        return the tokens and strings indicating the end of generation
        Args:
            eod_id (int): the end of document token id
            end_strings (List[str]): the list of end of generation strings
        Returns:
            a pair `(tokens, strings)` where `tokens` is a set of tokens (int) and `strings` is a list of strings,
            which must all be used to identify the end of generation (`tokens` always contains `eod_id`, while
            `strings` may be empty if all end strings are associated to unique tokens)
        """
        tokenizer = self.model.tokenizer
        # A cache is used to remember which end strings are associated to unique tokens vs. which ones
        # require an actual string comparison.
        if self._end_of_generation_cache is None or self._end_of_generation_cache["tokenizer"] is not tokenizer:
            # Invalidate the cache.
            self._end_of_generation_cache = {
                "tokenizer": tokenizer,
                "end_string_to_token": {END_OF_SEQ: eod_id},
                "end_strings_to_check": set(),
            }
        end_string_to_token = self._end_of_generation_cache["end_string_to_token"]

        end_tokens = {eod_id}  # always include `eod_id`, even if `END_OF_SEQ` is not within `end_strings`
        end_strings_to_check = []  # will contain end strings that have no associated special token

        for end_string in end_strings:
            try:
                end_tokens.add(end_string_to_token[end_string])
                continue
            except KeyError:
                if end_string in self._end_of_generation_cache["end_strings_to_check"]:
                    end_strings_to_check.append(end_string)
                    continue

            # `end_string` does not exist in the cache yet: check if `end_string` is a special token for
            # the tokenizer. Ideally, we would simply use `tokenizer.text_to_ids(end_string)`, but some
            # tokenizers (e.g., SentencePiece) may prefix the special token with another token associated
            # to an empty string. The code below is thus meant to extract the special token associated to
            # `end_string` (if it exists). Note that we use "<extra_id_1>" as prefix string to have a low
            # risk of the tokenizer merging it with `end_string`, but this is somewhat arbitrary.
            ids_ref = tokenizer.text_to_ids("<extra_id_1>")
            ids_with_end_string = tokenizer.text_to_ids(f"<extra_id_1>{end_string}")
            if len(ids_with_end_string) == len(ids_ref) + 1 and ids_with_end_string[:-1] == ids_ref:
                # We can assume that the extra token is the one corresponding to `end_string`.
                end_string_to_token[end_string] = ids_with_end_string[-1]
                end_tokens.add(ids_with_end_string[-1])
            else:
                # No special token.
                warnings.warn(
                    f"The end string '{end_string}' has no associated special token: this may slow down "
                    "generation (consider using a different tokenizer or modifying `end_strings`)"
                )
                self._end_of_generation_cache["end_strings_to_check"].add(end_string)
                end_strings_to_check.append(end_string)

        return end_tokens, end_strings_to_check


class SpeechToTextGenerationStrategy(TextGenerationStrategy):
    def __init__(self, model: pl.LightningModule, max_depth: int = 100):
        cnt = 0
        while not hasattr(model, "language_model") and cnt < max_depth:
            if not hasattr(model, "module"):
                raise ValueError(
                    f"Model does not have `language_model` or `module` attribute: {model.__class__.__name__}"
                )
            model = model.module
            cnt += 1
            if not hasattr(model, "language_model") and cnt == max_depth:
                raise ValueError(
                    f"Cannot find `language_model` attribute after {max_depth} iterations: {model.__class__.__name__}"
                )
        super().__init__(model)

    def init_batch(
        self,
        context_tokens: torch.Tensor,
        context_lengths: torch.Tensor,
        audio_signal: torch.Tensor,
        audio_length: torch.Tensor,
        compute_attention_mask: bool = True,
        num_audios: Optional[torch.Tensor] = None,
        context_start_idx: Optional[List[List[int]]] = None,
    ):
        """initialize the batch data before the inference steps."""

        audio_feats, audio_feat_lens = self.model.perception(
            input_signal=audio_signal,
            input_signal_length=audio_length,
            processed_signal=None,
            processed_signal_length=None,
        )

        if num_audios is not None:
            # handle multiple audio files per sample
            audio_feats = audio_feats.split(num_audios.tolist())
            audio_feat_lens = audio_feat_lens.split(num_audios.tolist())

        encoder_input, attention_mask, _, position_ids, encoder_max_length = self.model.inject_perception_input(
            audio_feats, audio_feat_lens, context_tokens, context_lengths, context_start_idx
        )

        self.attention_mask = attention_mask
        self.position_ids = position_ids

        if num_audios is not None:
            # handle multiple audio files per sample
            new_context_tokens = shift_tokens_by_multi_audios(
                context_tokens, context_lengths, audio_feat_lens, context_start_idx, encoder_max_length
            )
            audio_feat_lens = torch.stack([torch.sum(lens) for lens in audio_feat_lens])  # [batch,]
        else:
            new_context_tokens = self.model._shift_labels_by_emb_len(
                context_tokens, context_lengths, audio_feat_lens, encoder_max_length, pad_token=0
            )

        return new_context_tokens, encoder_input, audio_feat_lens

    def clip_max_len(self, maxlen: int) -> int:
        """clip the max len based on the LM model max sequence length"""
        # for positional embedding types that allow length extrapolation, don't clip the max length
        if self.model.language_model.position_embedding_type == "learned_absolute":
            if maxlen > self.model.language_model.max_sequence_length + 1:
                maxlen = self.model.language_model.max_sequence_length + 1
        return maxlen

    def prepare_batch_at_step(
        self,
        tokens: torch.Tensor,
        input_embeddings: torch.Tensor,
        maxlen: int,
        micro_batch_size: int,
        step: int,
        context_lengths: torch.Tensor,
        curr_context_length: int,
        compute_attention_mask: bool,
    ) -> Dict[str, torch.Tensor]:
        """Prepare batch for each of the inference steps"""
        if step == 0:
            tokens2use = tokens[:, :curr_context_length].contiguous()
            # positions2use = self.position_ids[:, :curr_context_length].contiguous()
            embeddings2use = input_embeddings[:curr_context_length].contiguous()
            # Prepare KV cache for the entire context at first step, and reuse afterwards.
            self.inference_params = InferenceParams(max_batch_size=tokens2use.size(0), max_sequence_length=maxlen)
        else:
            tokens2use = tokens[:, curr_context_length - 1].view(micro_batch_size, -1).contiguous()
            positions2use = self.position_ids[:, curr_context_length - 1].view(micro_batch_size, -1).contiguous()
            embeddings2use = self.model._get_text_embeddings(tokens2use, positions2use).contiguous()
            started = context_lengths <= curr_context_length
            embeddings2use = switch(input_embeddings[curr_context_length - 1].unsqueeze(0), embeddings2use, started)
            embeddings2use = embeddings2use.contiguous()

        attention_mask = self.attention_mask if compute_attention_mask else None
        batch = {
            "input_ids": None,
            "position_ids": None,
            "attention_mask": attention_mask,
            "decoder_input": embeddings2use,
            "inference_params": self.inference_params,
        }
        return batch

    def post_process(self, tokens: torch.Tensor, new_tokens: torch.Tensor, context_length: int):
        """
        At the end of the inference, post process the inference results
        """
        pass

    def end_of_generation_condition(
        self, tokens: torch.Tensor, prev: torch.Tensor, eod_id: int, end_strings: List[str]
    ) -> torch.Tensor:
        """
        return whether the generation should stop based on the previous token
        Args:
            tokens (torch.Tensor): the generated tokens so far
            prev  (torch.Tensor): the previous token
            eod_id (int): the end of document token id
            end_strings (List[str]): the list of end of generation strings
        returns:
            a boolean tensor indicating whether the generation should stop
        """
        if len(end_strings) == 1 and end_strings[0] == END_OF_SEQ:
            return prev == eod_id
        else:
            tokenizer = self.model.tokenizer
            conditions = []
            end_tokens = set()
            end_tokens.add(eod_id)
            for end_string in end_strings:
                if len(end_string) > 1:
                    continue
                ids_1 = tokenizer.text_to_ids(f'<extra_id_1>{end_string}')
                ids_2 = tokenizer.text_to_ids('<extra_id_1>')
                if len(ids_1) <= len(ids_2):
                    continue
                token_id = ids_1[len(ids_2) :][0]

                end_tokens.add(token_id)

            for p, token_item in zip(prev, tokens):
                text = tokenizer.ids_to_text(token_item.tolist())
                conditions.append(
                    any([text.endswith(end_string) for end_string in end_strings] + [p.item() in end_tokens])
                )
            return torch.tensor(conditions, dtype=torch.bool, device=tokens.device)


def model_inference_strategy_dispatcher(model, **args):
    from nemo.collections.speechlm.models.speech_to_text_llm_model import SpeechToTextLLM

    if isinstance(model, SpeechToTextLLM):
        return SpeechToTextGenerationStrategy(model, **args)
    else:
        raise ValueError(f"Unsupported model type for text generation: {model.__class__.__name__}")



# Copyright (c) 2022, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from functools import reduce
from typing import List

import torch
import torch.nn as nn


class FusedBatchNorm1d(nn.Module):
    """
    Fused BatchNorm to use in Conformer to improve accuracy in finetuning with TTS scenario
    Drop-in replacement for BatchNorm1d with simple affine projection
    """

    def __init__(self, num_features: int):
        """
        Args:
            num_features: number of channels, see original BatchNorm1d documentation
        """
        super().__init__()
        self.num_features = num_features
        self.weight = nn.Parameter(torch.ones(num_features))
        self.bias = nn.Parameter(torch.zeros(num_features))

    @classmethod
    def from_batchnorm(cls, bn: nn.BatchNorm1d) -> FusedBatchNorm1d:
        """
        Construct FusedBatchNorm1d module from BatchNorm1d
        Args:
            bn: original BatchNorm module

        Returns:
            FusedBatchNorm1d module with initialized params; in eval mode result is equivalent to original BatchNorm
        """
        assert isinstance(bn, nn.BatchNorm1d)
        fused_bn = FusedBatchNorm1d(bn.num_features)
        # init projection params from original batch norm
        # so, for inference mode output is the same
        std = torch.sqrt(bn.running_var.data + bn.eps)
        fused_bn.weight.data = bn.weight.data / std
        fused_bn.bias.data = bn.bias.data - bn.running_mean.data * fused_bn.weight.data
        return fused_bn

    def forward(self, x: torch.Tensor):
        if x.dim() == 3:
            return x * self.weight.unsqueeze(-1) + self.bias.unsqueeze(-1)
        assert x.dim() == 2
        return x * self.weight + self.bias


def _get_module_by_name(module: nn.Module, full_layer_name: str) -> nn.Module:
    names = full_layer_name.split(sep='.')
    return reduce(getattr, names, module)


def replace_bn_with_fused_bn(module: nn.Module, full_layer_name: str):
    """
    Replace BatchNorm1d named `full_layer_name` in nn.Module with FusedBatchNorm1d
    Args:
        module: nn.Module instance, modified inplace
        full_layer_name: name of BatchNorm1d submodule in module to replace
    """
    bn = _get_module_by_name(module, full_layer_name)
    assert isinstance(bn, nn.BatchNorm1d)
    fused_bn = FusedBatchNorm1d.from_batchnorm(bn)
    try:
        parent_name, norm_name = full_layer_name.rsplit(".", maxsplit=1)
        setattr(_get_module_by_name(module, parent_name), norm_name, fused_bn)
    except ValueError:
        norm_name = full_layer_name
        setattr(module, norm_name, fused_bn)


def replace_bn_with_fused_bn_all(model: nn.Module) -> List[str]:
    """
    Replace BatchNorm1d with FusedBatchNorm1d in model
    Args:
        model: nn.Module instance, modified inplace

    Returns:
        list of replaced module names
    """
    replaced_module_names = []
    for name, module in model.named_modules():
        if isinstance(module, nn.BatchNorm1d):
            replace_bn_with_fused_bn(model, name)
            replaced_module_names.append(name)
    return replaced_module_names



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2018-2020 William Falcon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

import torch

from .pl_utils import BATCH_SIZE, NUM_BATCHES, NUM_CLASSES


@dataclass(frozen=True)
class LossInput:
    """
    The input for ``nemo.collections.common.metrics.GlobalAverageLossMetric`` metric tests.

    Args:
        loss_sum_or_avg: a one dimensional float tensor which contains losses for averaging. Each element is either a
            sum or mean of several losses depending on the parameter ``take_avg_loss`` of the
            ``nemo.collections.common.metrics.GlobalAverageLossMetric`` class.
        num_measurements: a one dimensional integer tensor which contains number of measurements which sums or average
            values are in ``loss_sum_or_avg``.
    """

    loss_sum_or_avg: torch.Tensor
    num_measurements: torch.Tensor


NO_ZERO_NUM_MEASUREMENTS = LossInput(
    loss_sum_or_avg=torch.rand(NUM_BATCHES) * 2.0 - 1.0, num_measurements=torch.randint(1, 100, (NUM_BATCHES,)),
)

SOME_NUM_MEASUREMENTS_ARE_ZERO = LossInput(
    loss_sum_or_avg=torch.rand(NUM_BATCHES) * 2.0 - 1.0,
    num_measurements=torch.cat(
        (
            torch.randint(1, 100, (NUM_BATCHES // 2,), dtype=torch.int32),
            torch.zeros(NUM_BATCHES - NUM_BATCHES // 2, dtype=torch.int32),
        )
    ),
)

ALL_NUM_MEASUREMENTS_ARE_ZERO = LossInput(
    loss_sum_or_avg=torch.rand(NUM_BATCHES) * 2.0 - 1.0, num_measurements=torch.zeros(NUM_BATCHES, dtype=torch.int32),
)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import torch
import torch.nn as nn
from torch.nn import LayerNorm

from nemo.collections.asr.parts.submodules.causal_convs import CausalConv1D, CausalConv2D
from nemo.utils import logging


class StackingSubsampling(torch.nn.Module):
    """Stacking subsampling which simply stacks consecutive frames to reduce the sampling rate
    Args:
        subsampling_factor (int): The subsampling factor
        feat_in (int): size of the input features
        feat_out (int): size of the output features
        norm (bool): whether to use an MLP layer after the stacking along with normalization. default is False.
    """

    def __init__(self, subsampling_factor, feat_in, feat_out, norm=False):
        super(StackingSubsampling, self).__init__()
        self.subsampling_factor = subsampling_factor
        self.proj_out = torch.nn.Linear(subsampling_factor * feat_in, feat_out)
        if norm:
            self.pre_norm = LayerNorm(feat_in)
        else:
            self.pre_norm = None

    def get_sampling_frames(self):
        return self.subsampling_factor

    def get_streaming_cache_size(self):
        return 0

    def forward(self, x, lengths):
        b, t, h = x.size()
        pad_size = (self.subsampling_factor - (t % self.subsampling_factor)) % self.subsampling_factor
        x = torch.nn.functional.pad(x, (0, 0, 0, pad_size))
        if self.pre_norm is not None:
            x = self.pre_norm(x)
        _, t, _ = x.size()
        x = torch.reshape(x, (b, t // self.subsampling_factor, h * self.subsampling_factor))
        x = self.proj_out(x)
        lengths = torch.div(lengths + pad_size, self.subsampling_factor, rounding_mode='floor')
        return x, lengths


class ConvSubsampling(torch.nn.Module):
    """Convolutional subsampling which supports VGGNet and striding approach introduced in:
    VGGNet Subsampling: Transformer-transducer: end-to-end speech recognition with self-attention (https://arxiv.org/pdf/1910.12977.pdf)
    Striding Subsampling: "Speech-Transformer: A No-Recurrence Sequence-to-Sequence Model for Speech Recognition" by Linhao Dong et al. (https://ieeexplore.ieee.org/document/8462506)
    Args:
        subsampling (str): The subsampling technique from {"vggnet", "striding", "dw-striding"}
        subsampling_factor (int): The subsampling factor which should be a power of 2
        subsampling_conv_chunking_factor (int): Input chunking factor which can be -1 (no chunking)
        1 (auto) or a power of 2. Default is 1
        feat_in (int): size of the input features
        feat_out (int): size of the output features
        conv_channels (int): Number of channels for the convolution layers.
        activation (Module): activation function, default is nn.ReLU()
    """

    def __init__(
        self,
        subsampling,
        subsampling_factor,
        feat_in,
        feat_out,
        conv_channels,
        subsampling_conv_chunking_factor=1,
        activation=nn.ReLU(),
        is_causal=False,
    ):
        super(ConvSubsampling, self).__init__()
        self._subsampling = subsampling
        self._conv_channels = conv_channels
        self._feat_in = feat_in
        self._feat_out = feat_out

        if subsampling_factor % 2 != 0:
            raise ValueError("Sampling factor should be a multiply of 2!")
        self._sampling_num = int(math.log(subsampling_factor, 2))
        self.subsampling_factor = subsampling_factor
        self.is_causal = is_causal

        if (
            subsampling_conv_chunking_factor != -1
            and subsampling_conv_chunking_factor != 1
            and subsampling_conv_chunking_factor % 2 != 0
        ):
            raise ValueError("subsampling_conv_chunking_factor should be -1, 1, or a power of 2")
        self.subsampling_conv_chunking_factor = subsampling_conv_chunking_factor

        in_channels = 1
        layers = []

        if subsampling == 'vggnet':
            self._stride = 2
            self._kernel_size = 2
            self._ceil_mode = True

            self._left_padding = 0
            self._right_padding = 0

            for i in range(self._sampling_num):
                layers.append(
                    torch.nn.Conv2d(
                        in_channels=in_channels, out_channels=conv_channels, kernel_size=3, stride=1, padding=1
                    )
                )
                layers.append(activation)
                layers.append(
                    torch.nn.Conv2d(
                        in_channels=conv_channels, out_channels=conv_channels, kernel_size=3, stride=1, padding=1
                    )
                )
                layers.append(activation)
                layers.append(
                    torch.nn.MaxPool2d(
                        kernel_size=self._kernel_size,
                        stride=self._stride,
                        padding=self._left_padding,
                        ceil_mode=self._ceil_mode,
                    )
                )
                in_channels = conv_channels

        elif subsampling == 'dw_striding':
            self._stride = 2
            self._kernel_size = 3
            self._ceil_mode = False

            if self.is_causal:
                self._left_padding = self._kernel_size - 1
                self._right_padding = self._stride - 1
                self._max_cache_len = subsampling_factor + 1
            else:
                self._left_padding = (self._kernel_size - 1) // 2
                self._right_padding = (self._kernel_size - 1) // 2
                self._max_cache_len = 0

            # Layer 1
            if self.is_causal:
                layers.append(
                    CausalConv2D(
                        in_channels=in_channels,
                        out_channels=conv_channels,
                        kernel_size=self._kernel_size,
                        stride=self._stride,
                        padding=None,
                    )
                )
            else:
                layers.append(
                    torch.nn.Conv2d(
                        in_channels=in_channels,
                        out_channels=conv_channels,
                        kernel_size=self._kernel_size,
                        stride=self._stride,
                        padding=self._left_padding,
                    )
                )
            in_channels = conv_channels
            layers.append(activation)

            for i in range(self._sampling_num - 1):
                if self.is_causal:
                    layers.append(
                        CausalConv2D(
                            in_channels=in_channels,
                            out_channels=in_channels,
                            kernel_size=self._kernel_size,
                            stride=self._stride,
                            padding=None,
                            groups=in_channels,
                        )
                    )
                else:
                    layers.append(
                        torch.nn.Conv2d(
                            in_channels=in_channels,
                            out_channels=in_channels,
                            kernel_size=self._kernel_size,
                            stride=self._stride,
                            padding=self._left_padding,
                            groups=in_channels,
                        )
                    )

                layers.append(
                    torch.nn.Conv2d(
                        in_channels=in_channels,
                        out_channels=conv_channels,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        groups=1,
                    )
                )
                layers.append(activation)
                in_channels = conv_channels

        elif subsampling == 'striding':
            self._stride = 2
            self._kernel_size = 3
            self._ceil_mode = False

            if self.is_causal:
                self._left_padding = self._kernel_size - 1
                self._right_padding = self._stride - 1
                self._max_cache_len = subsampling_factor + 1
            else:
                self._left_padding = (self._kernel_size - 1) // 2
                self._right_padding = (self._kernel_size - 1) // 2
                self._max_cache_len = 0

            for i in range(self._sampling_num):
                if self.is_causal:
                    layers.append(
                        CausalConv2D(
                            in_channels=in_channels,
                            out_channels=conv_channels,
                            kernel_size=self._kernel_size,
                            stride=self._stride,
                            padding=None,
                        )
                    )
                else:
                    layers.append(
                        torch.nn.Conv2d(
                            in_channels=in_channels,
                            out_channels=conv_channels,
                            kernel_size=self._kernel_size,
                            stride=self._stride,
                            padding=self._left_padding,
                        )
                    )
                layers.append(activation)
                in_channels = conv_channels

        elif subsampling == 'striding_conv1d':

            in_channels = feat_in

            self._stride = 2
            self._kernel_size = 5
            self._ceil_mode = False

            if self.is_causal:
                self._left_padding = self._kernel_size - 1
                self._right_padding = self._stride - 1
                self._max_cache_len = subsampling_factor + 1
            else:
                self._left_padding = (self._kernel_size - 1) // 2
                self._right_padding = (self._kernel_size - 1) // 2
                self._max_cache_len = 0

            for i in range(self._sampling_num):
                if self.is_causal:
                    layers.append(
                        CausalConv1D(
                            in_channels=in_channels,
                            out_channels=feat_out if self._sampling_num == i + 1 else conv_channels,
                            kernel_size=self._kernel_size,
                            stride=self._stride,
                            padding=None,
                        )
                    )
                else:
                    layers.append(
                        torch.nn.Conv1d(
                            in_channels=in_channels,
                            out_channels=feat_out if self._sampling_num == i + 1 else conv_channels,
                            kernel_size=self._kernel_size,
                            stride=self._stride,
                            padding=self._left_padding,
                        )
                    )
                layers.append(activation)
                in_channels = conv_channels

        elif subsampling == 'dw_striding_conv1d':

            in_channels = feat_in

            self._stride = 2
            self._kernel_size = 5
            self._ceil_mode = False

            self._left_padding = (self._kernel_size - 1) // 2
            self._right_padding = (self._kernel_size - 1) // 2

            # Layer 1
            layers.extend(
                [
                    torch.nn.Conv1d(
                        in_channels=in_channels,
                        out_channels=in_channels,
                        kernel_size=self._kernel_size,
                        stride=self._stride,
                        padding=self._left_padding,
                        groups=in_channels,
                    ),
                    torch.nn.Conv1d(
                        in_channels=in_channels,
                        out_channels=feat_out if self._sampling_num == 1 else conv_channels,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        groups=1,
                    ),
                ]
            )
            in_channels = conv_channels
            layers.append(activation)

            for i in range(self._sampling_num - 1):
                layers.extend(
                    [
                        torch.nn.Conv1d(
                            in_channels=in_channels,
                            out_channels=in_channels,
                            kernel_size=self._kernel_size,
                            stride=self._stride,
                            padding=self._left_padding,
                            groups=in_channels,
                        ),
                        torch.nn.Conv1d(
                            in_channels=in_channels,
                            out_channels=feat_out if self._sampling_num == i + 2 else conv_channels,
                            kernel_size=1,
                            stride=1,
                            padding=0,
                            groups=1,
                        ),
                    ]
                )
                layers.append(activation)
                in_channels = conv_channels

        else:
            raise ValueError(f"Not valid sub-sampling: {subsampling}!")

        if subsampling in ["vggnet", "dw_striding", "striding"]:

            in_length = torch.tensor(feat_in, dtype=torch.float)
            out_length = calc_length(
                lengths=in_length,
                all_paddings=self._left_padding + self._right_padding,
                kernel_size=self._kernel_size,
                stride=self._stride,
                ceil_mode=self._ceil_mode,
                repeat_num=self._sampling_num,
            )
            self.out = torch.nn.Linear(conv_channels * int(out_length), feat_out)
            self.conv2d_subsampling = True
        elif subsampling in ["striding_conv1d", "dw_striding_conv1d"]:
            self.out = None
            self.conv2d_subsampling = False
        else:
            raise ValueError(f"Not valid sub-sampling: {subsampling}!")

        self.conv = MaskedConvSequential(*layers)

    def get_sampling_frames(self):
        return [1, self.subsampling_factor]

    def get_streaming_cache_size(self):
        return [0, self.subsampling_factor + 1]

    def forward(self, x, lengths):
        out_lengths = calc_length(
            lengths,
            all_paddings=self._left_padding + self._right_padding,
            kernel_size=self._kernel_size,
            stride=self._stride,
            ceil_mode=self._ceil_mode,
            repeat_num=self._sampling_num,
        )

        # Transpose to Channel First mode
        if not self.conv2d_subsampling:
            x = x.transpose(1, 2)

        # split inputs if chunking_factor is set
        if self.subsampling_conv_chunking_factor != -1 and self.conv2d_subsampling:
            if self.subsampling_conv_chunking_factor == 1:
                # if subsampling_conv_chunking_factor is 1, we split only if needed
                # avoiding a bug / feature limiting indexing of tensors to 2**31
                # see https://github.com/pytorch/pytorch/issues/80020
                x_ceil = 2**31 / self._conv_channels * self._stride * self._stride
                if torch.numel(x) > x_ceil:
                    need_to_split = True
                else:
                    need_to_split = False
            else:
                # if subsampling_conv_chunking_factor > 1 we always split
                need_to_split = True

            if need_to_split:
                x, lengths, success = self.conv_split_by_batch(x, lengths)
                if not success:  # if unable to split by batch, try by channel
                    if self._subsampling == 'dw_striding':
                        # TODO: implement lengths inside conv_split_by_channel
                        x = self.conv_split_by_channel(x)
                        lengths = out_lengths
                    else:
                        x, lengths = self.conv(x, lengths)  # try anyway
            else:
                x, lengths = self.conv(x, lengths)
        else:
            x, lengths = self.conv(x)

        # Flatten Channel and Frequency Axes
        if self.conv2d_subsampling:
            b, c, t, f = x.size()
            x = self.out(x.transpose(1, 2).reshape(b, t, -1))
        # Transpose to Channel Last mode
        else:
            x = x.transpose(1, 2)

        return x, lengths

    def reset_parameters(self):
        # initialize weights
        if self._subsampling == 'dw_striding':
            with torch.no_grad():
                # init conv
                scale = 1.0 / self._kernel_size
                dw_max = (self._kernel_size**2) ** -0.5
                pw_max = self._conv_channels**-0.5

                torch.nn.init.uniform_(self.conv[0].weight, -scale, scale)
                torch.nn.init.uniform_(self.conv[0].bias, -scale, scale)

                for idx in range(2, len(self.conv), 3):
                    torch.nn.init.uniform_(self.conv[idx].weight, -dw_max, dw_max)
                    torch.nn.init.uniform_(self.conv[idx].bias, -dw_max, dw_max)
                    torch.nn.init.uniform_(self.conv[idx + 1].weight, -pw_max, pw_max)
                    torch.nn.init.uniform_(self.conv[idx + 1].bias, -pw_max, pw_max)

                # init fc (80 * 64 = 5120 from https://github.com/kssteven418/Squeezeformer/blob/13c97d6cf92f2844d2cb3142b4c5bfa9ad1a8951/src/models/conformer_encoder.py#L487
                fc_scale = (self._feat_out * self._feat_in / self._sampling_num) ** -0.5
                torch.nn.init.uniform_(self.out.weight, -fc_scale, fc_scale)
                torch.nn.init.uniform_(self.out.bias, -fc_scale, fc_scale)

    def conv_split_by_batch(self, x, lengths):
        """Tries to split input by batch, run conv and concat results"""
        b, *_ = x.size()
        if b == 1:  # can't split if batch size is 1
            return x, lengths, False

        if self.subsampling_conv_chunking_factor > 1:
            cf = self.subsampling_conv_chunking_factor
            logging.debug(f'using manually set chunking factor: {cf}')
        else:
            # avoiding a bug / feature limiting indexing of tensors to 2**31
            # see https://github.com/pytorch/pytorch/issues/80020
            x_ceil = 2**31 / self._conv_channels * self._stride * self._stride
            p = math.ceil(math.log(torch.numel(x) / x_ceil, 2))
            cf = 2**p
            logging.debug(f'using auto set chunking factor: {cf}')

        new_batch_size = b // cf
        if new_batch_size == 0:  # input is too big
            return x, lengths, False

        logging.debug(f'conv subsampling: using split batch size {new_batch_size}')

        ans = [
            self.conv(chunk, ln)
            for chunk, ln in zip(
                torch.split(x, new_batch_size, 0),
                torch.split(lengths, new_batch_size, 0),
            )
        ]
        return torch.cat([a[0] for a in ans]), torch.cat([a[1] for a in ans]), True

    def conv_split_by_channel(self, x):
        """For dw convs, tries to split input by time, run conv and concat results"""

        # Note: this method doesn't use the convolution masking implemented in MaskedConvolutionSequential
        x = x.unsqueeze(0)
        x = self.conv[0](x)  # full conv2D
        x = self.conv[1](x)  # activation

        for i in range(self._sampling_num - 1):
            _, c, t, _ = x.size()

            if self.subsampling_conv_chunking_factor > 1:
                cf = self.subsampling_conv_chunking_factor
                logging.debug(f'using manually set chunking factor: {cf}')
            else:
                # avoiding a bug / feature limiting indexing of tensors to 2**31
                # see https://github.com/pytorch/pytorch/issues/80020
                p = math.ceil(math.log(torch.numel(x) / 2**31, 2))
                cf = 2**p
                logging.debug(f'using auto set chunking factor: {cf}')

            new_c = int(c // cf)
            if new_c == 0:
                logging.warning(f'chunking factor {cf} is too high; splitting down to one channel.')
                new_c = 1

            new_t = int(t // cf)
            if new_t == 0:
                logging.warning(f'chunking factor {cf} is too high; splitting down to one timestep.')
                new_t = 1

            logging.debug(f'conv dw subsampling: using split C size {new_c} and split T size {new_t}')
            x = self.channel_chunked_conv(self.conv[i * 3 + 2], new_c, x)  # conv2D, depthwise

            # splitting pointwise convs by time
            x = torch.cat([self.conv[i * 3 + 3](chunk) for chunk in torch.split(x, new_t, 2)], 2)  # conv2D, pointwise
            x = self.conv[i * 3 + 4](x)  # activation
        return x

    def channel_chunked_conv(self, conv, chunk_size, x):
        """Performs channel chunked convolution"""

        ind = 0
        out_chunks = []
        for chunk in torch.split(x, chunk_size, 1):
            step = chunk.size()[1]

            if self.is_causal:
                chunk = nn.functional.pad(
                    chunk, pad=(self._kernel_size - 1, self._stride - 1, self._kernel_size - 1, self._stride - 1)
                )
                ch_out = nn.functional.conv2d(
                    chunk,
                    conv.weight[ind : ind + step, :, :, :],
                    bias=conv.bias[ind : ind + step],
                    stride=self._stride,
                    padding=0,
                    groups=step,
                )
            else:
                ch_out = nn.functional.conv2d(
                    chunk,
                    conv.weight[ind : ind + step, :, :, :],
                    bias=conv.bias[ind : ind + step],
                    stride=self._stride,
                    padding=self._left_padding,
                    groups=step,
                )
            out_chunks.append(ch_out)
            ind += step

        return torch.cat(out_chunks, 1)

    def change_subsampling_conv_chunking_factor(self, subsampling_conv_chunking_factor: int):
        if (
            subsampling_conv_chunking_factor != -1
            and subsampling_conv_chunking_factor != 1
            and subsampling_conv_chunking_factor % 2 != 0
        ):
            raise ValueError("subsampling_conv_chunking_factor should be -1, 1, or a power of 2")
        self.subsampling_conv_chunking_factor = subsampling_conv_chunking_factor


def calc_length(lengths, all_paddings, kernel_size, stride, ceil_mode, repeat_num=1):
    """Calculates the output length of a Tensor passed through a convolution or max pooling layer"""
    add_pad: float = all_paddings - kernel_size
    one: float = 1.0
    for i in range(repeat_num):
        lengths = torch.div(lengths.to(dtype=torch.float) + add_pad, stride) + one
        if ceil_mode:
            lengths = torch.ceil(lengths)
        else:
            lengths = torch.floor(lengths)
    return lengths.to(dtype=torch.int)


class TimeReductionModule(nn.Module):
    """
    Squeezeformer Time Reduction procedure. Downsamples the audio by `stride` in the time dimension.

    Args:
        d_model (int): input dimension of MultiheadAttentionMechanism and PositionwiseFeedForward
        out_dim (int): Output dimension of the module.
        kernel_size (int): Conv kernel size for depthwise convolution in convolution module
        stride (int): Downsampling factor in time dimension.
    """

    def __init__(self, d_model: int, out_dim: int, kernel_size: int = 5, stride: int = 2):
        super().__init__()

        self.d_model = d_model
        self.out_dim = out_dim
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = max(0, self.kernel_size - self.stride)

        self.dw_conv = nn.Conv1d(
            in_channels=d_model,
            out_channels=d_model,
            kernel_size=kernel_size,
            stride=stride,
            padding=self.padding,
            groups=d_model,
        )

        self.pw_conv = nn.Conv1d(
            in_channels=d_model,
            out_channels=out_dim,
            kernel_size=1,
            stride=1,
            padding=0,
            groups=1,
        )

        self.reset_parameters()

    def forward(self, x, att_mask=None, pad_mask=None):
        x = x.transpose(1, 2)  # [B, C, T]
        if pad_mask is not None:
            x = x.float().masked_fill(pad_mask.unsqueeze(1), 0.0)

        x = self.dw_conv(x)
        x = self.pw_conv(x)

        x = x.transpose(1, 2)  # [B, T, C]

        B, T, D = x.size()
        if att_mask is not None and pad_mask is not None:
            att_mask = att_mask[:, :: self.stride, :: self.stride]
            pad_mask = pad_mask[:, :: self.stride]
            L = pad_mask.size(-1)
            x = torch.nn.functional.pad(x, (0, 0, 0, L - T))

        return x, att_mask, pad_mask

    def reset_parameters(self):
        dw_max = self.kernel_size**-0.5
        pw_max = self.d_model**-0.5

        with torch.no_grad():
            torch.nn.init.uniform_(self.dw_conv.weight, -dw_max, dw_max)
            torch.nn.init.uniform_(self.dw_conv.bias, -dw_max, dw_max)
            torch.nn.init.uniform_(self.pw_conv.weight, -pw_max, pw_max)
            torch.nn.init.uniform_(self.pw_conv.bias, -pw_max, pw_max)


class SubsamplingReductionModule(nn.Module):
    """Downsamples the audio signal in time dimension."""

    def __init__(self, reduction: str, d_model: int, reduction_factor: int = 2):
        super().__init__()

        assert reduction in ['pooling', 'striding']

        self.reduction = reduction
        self.d_model = d_model
        self._sampling_num = int(math.log(reduction_factor, 2))

        if reduction == 'pooling':
            self.reduction_enc = nn.MaxPool1d(kernel_size=reduction_factor)
            self.padding = 0
            self.kernel_size = self.reduction_enc.kernel_size
            self.stride = self.reduction_enc.stride
        elif reduction == 'striding':
            self.reduction_enc = ConvSubsampling(
                subsampling='striding',
                subsampling_factor=reduction_factor,
                feat_in=d_model,
                feat_out=d_model,
                conv_channels=d_model,
                activation=nn.ReLU(),
                is_causal=False,
            )

    def forward(self, x, lengths):
        """Shapes:
        - x: [B, T, C]
        - lengths: [B]
        """

        if self.reduction == 'striding':
            x, lengths = self.reduction_enc(x=x, lengths=lengths)
        else:
            x = torch.transpose(x, 1, 2)  # [B, C, T]
            lengths = calc_length(
                lengths=lengths,
                all_paddings=self.padding,
                kernel_size=self.kernel_size,
                stride=self.stride,
                ceil_mode=False,
                repeat_num=self._sampling_num,
            )
            x = self.reduction_enc(x)
            x = torch.transpose(x, 1, 2)  # [B, T, C]

        return x, lengths


def apply_channel_mask(tensor, mask):
    """Apply mask to tensor with channel dimension."""
    # tensor: (batch, channels, time, features)
    # mask: (batch, time, features)
    batch_size, channels, time, features = tensor.shape
    expanded_mask = mask.unsqueeze(1).expand(batch_size, channels, time, features)
    return tensor * expanded_mask


def calculate_conv_output_size(input_size: torch.Tensor, kernel_size: int, stride: int, padding: tuple[int, int]):
    """Calculate exact output size after convolution."""
    return (input_size + padding[0] + padding[1] - kernel_size) // stride + 1


class MaskedConvSequential(nn.Sequential):
    def forward(self, x, lengths):
        # Convert input (batch, time, features) to conv format
        x = x.unsqueeze(1)  # (batch, 1, time, features)
        current_lengths = lengths.clone().float()
        mask = self._create_mask(x, current_lengths.long())

        # Process through each layer with mask propagation
        for i, layer in enumerate(self):
            # Apply current mask before layer
            x = apply_channel_mask(x, mask)

            # Apply layer
            x = layer(x)

            # Update lengths for stride operations with proper padding
            if hasattr(layer, 'stride') and layer.stride != (1, 1):
                if hasattr(layer, "_left_padding"):
                    padding = (layer._left_padding, layer._right_padding)  # CausalConv2D
                else:
                    padding = layer.padding
                current_lengths = calculate_conv_output_size(
                    current_lengths, layer.kernel_size[0], layer.stride[0], padding
                )
                mask = self._create_mask(x, current_lengths.long())

        # Final masking
        x = apply_channel_mask(x, mask)
        return x, current_lengths.long()

    def _create_mask(self, tensor, lengths):
        """Create mask matching tensor dimensions."""
        batch_size, channels, time, features = tensor.shape
        time_mask = torch.arange(time, device=tensor.device).expand(batch_size, time) < lengths.unsqueeze(1)
        return time_mask.unsqueeze(-1).expand(batch_size, time, features).to(tensor.dtype)
